#include <stdio.h>
#include <dmusici.h>
#include <crtdbg.h>
#include <assert.h>
#define __BYTEBOOL__
#include "h2def.h"

typedef struct _MUSFile
{
	char	m_rgbID[4];
	WORD	m_wcbMusicData;
	WORD	m_wcbHeaderData;
	WORD	m_wcPrimaryChannels;
	WORD	m_wcSecondaryChannels;
	WORD	m_wcPatches;
	WORD	m_wReserved;
	WORD *	m_rgwPatches;
	BYTE *	m_rgbMusicData;
} MUSFile;

#pragma pack(push, 1)

typedef struct __MUSEvent
{
	BYTE	m_bChannel	: 4;
	BYTE	m_bEventType: 3;
	BYTE	m_bLast		: 1;
} _MUSEvent;

typedef union MUSEvent
{
	BYTE		m_bEvent;
	_MUSEvent	m_event;
} MUSEvent;

#pragma pack(pop)

typedef struct _MUS2MIDIState
{
	MUSFile			m_mf;
	BYTE *			m_pCurMUSEvent;
	BYTE *			m_rgmidEvents;
	BYTE *			m_pCurMIDEvent;
	DWORD			m_dwcMIDIBufferLength;
	BYTE			m_rgbChannelVolume[16];
	LONG			m_lNextWaitTime;
#ifdef _DEBUG
	char			m_szCurrentOperation[256];
#endif
} MUS2MIDIState;

typedef struct _SegmentToPointerMap
{
	// Note this pointer will never hold a reference!
	// Don't call Release() on it!
	//
	IDirectMusicSegment *	m_pDMSeg;
	byte *					m_rgbMIDIData;
} SegmentToPointerMap;


static void HandleKeyUp(MUSEvent *pevt, MUS2MIDIState *pM2M);
static void HandleKeyDown(MUSEvent *pevt, MUS2MIDIState *pM2M);
static void HandlePitchBend(MUSEvent *pevt, MUS2MIDIState *pM2M);
static void HandleSystemEvent(MUSEvent *pevt, MUS2MIDIState *pM2M);
static void HandleControllerEvent(MUSEvent *pevt, MUS2MIDIState *pM2M);
static void ConvertMUSEventsToMIDIEvents(MUS2MIDIState *pM2M);

static boolean g_fPlayMusic = true;
static unsigned char g_MIDIHeader[] = { 'M', 'T', 'h', 'd', 0, 0, 0, 6, 0, 0, 0, 1, 0, 0x59,
										'M', 'T', 'r', 'k', 0, 0, 0, 0,
//										0x00, 0xF0, 0x7E, 0x7F, 0x09, 0x01, 0xF7,
										0x00, 0xFF, 0x59, 0x02, 0x00, 0x00,
										0x00, 0xFF, 0x51, 0x03, 0x09, 0xA3, 0x1A };
static IDirectMusicPerformance *g_pDMP = NULL;
static IDirectMusicLoader *g_pDML = NULL;
static SegmentToPointerMap g_rgS2P[32];

//static MUSIC_TIME mtOffset = 0;
//static MUSIC_TIME mtStart = 0;

/*
 *
 *                              SONG API
 *
 */


void I_ShutdownMusic(void)
{
	if(g_pDMP)
	{
		g_pDMP->lpVtbl->Stop(g_pDMP, NULL, NULL, 0, 0);
		g_pDMP->lpVtbl->CloseDown(g_pDMP);
		g_pDMP->lpVtbl->Release(g_pDMP);
		g_pDMP = NULL;
	}
	if(g_pDML)
	{
		g_pDML->lpVtbl->ClearCache(g_pDML, &CLSID_DirectMusicSegment);
		g_pDML->lpVtbl->Release(g_pDML);
		g_pDML = NULL;
	}

	CoUninitialize();
}


HRESULT I_InitMusic(IDirectSound *pDS)
{
	HRESULT hr = S_OK;

	ZeroMemory(g_rgS2P, 32 * sizeof(SegmentToPointerMap));

	if(FAILED(hr = CoInitialize(NULL)))
	{
		return hr;
	}

	if(FAILED(hr = CoCreateInstance(&CLSID_DirectMusicPerformance,
									NULL,            
									CLSCTX_INPROC, 
									&IID_IDirectMusicPerformance,            
									(void**)&g_pDMP)))
	{
		g_fPlayMusic = false;
		return hr;
	}

	if(FAILED(hr = g_pDMP->lpVtbl->Init(g_pDMP, NULL, pDS, NULL)))
	{
		g_pDMP->lpVtbl->Release(g_pDMP);
		g_pDMP = NULL;
		g_fPlayMusic = false;
		return hr;
	}

	if(FAILED(hr = g_pDMP->lpVtbl->AddPort(g_pDMP, NULL)))
	{
		g_pDMP->lpVtbl->Release(g_pDMP);
		g_pDMP = NULL;
		g_fPlayMusic = false;
		return hr;
	}

	if(FAILED(hr = CoCreateInstance(&CLSID_DirectMusicLoader,
									NULL,
									CLSCTX_INPROC_SERVER,
									&IID_IDirectMusicLoader,
									(void **)&g_pDML)))
	{
		g_pDMP->lpVtbl->Release(g_pDMP);
		g_pDMP = NULL;
		g_fPlayMusic = false;
		return hr;
	}

	if(FAILED(hr = g_pDML->lpVtbl->EnableCache(g_pDML, &CLSID_DirectMusicSegment, FALSE)))
	{
		g_pDMP->lpVtbl->Release(g_pDMP);
		g_pDMP = NULL;
		g_pDML->lpVtbl->Release(g_pDML);
		g_pDML = NULL;
		g_fPlayMusic = false;
		return hr;
	}

/*
	if(FAILED(hr = g_pDMP->lpVtbl->SetPrepareTime(g_pDMP, 100)))
	{
		g_pDMP->lpVtbl->Release(g_pDMP);
		g_pDMP = NULL;
		g_pDML->lpVtbl->Release(g_pDML);
		g_pDML = NULL;
		g_fPlayMusic = false;
		return hr;
	}
*/

	g_fPlayMusic = true;
	return S_OK;
}


int I_QrySongPlaying(int handle)
{
	HRESULT hr;
	IDirectMusicSegment *pSeg;

	if(!g_fPlayMusic || !handle || !g_pDMP)
	{
		return 0;
	}

	pSeg = (IDirectMusicSegment *)handle;

	hr = g_pDMP->lpVtbl->IsPlaying(g_pDMP, pSeg, NULL);

	return hr == S_OK ? 1 : 0;

/*
	int rc = MUS_QrySongPlaying(handle);
#ifdef SNDDEBUG
  if (rc < 0) ST_Message("    MUS_QrySP() returned %d\n", rc);
#endif
  return rc;
*/
}


void I_PauseSong(int handle)
{
	IDirectMusicSegment *pSeg;
	MUSIC_TIME mt;

	if(!g_fPlayMusic || !handle || !g_pDMP)
	{
		return;
	}

	pSeg = (IDirectMusicSegment *)handle;

	pSeg->lpVtbl->GetStartPoint(pSeg, &mt);

	g_pDMP->lpVtbl->Stop(g_pDMP, pSeg, NULL, 0, 0);

	pSeg->lpVtbl->GetStartPoint(pSeg, &mt);

//  MUS_PauseSong(handle);
}

void I_ResumeSong(int handle)
{
	IDirectMusicSegment *pSeg;

	if(!g_fPlayMusic || !handle || !g_pDMP)
	{
		return;
	}

	pSeg = (IDirectMusicSegment *)handle;

	g_pDMP->lpVtbl->PlaySegment(g_pDMP, pSeg, 0, 0, NULL);

//  MUS_ResumeSong(handle);
}

void I_SetMusicVolume(int volume)
{
	if(!g_fPlayMusic || !g_pDMP)
	{
		return;
	}

	snd_MusicVolume = volume;
	volume = (15 - volume) * -100;

	g_pDMP->lpVtbl->SetGlobalParam(g_pDMP, &GUID_PerfMasterVolume, &volume, sizeof(long));

//  MUS_SetMasterVolume(volume*8);
}


int I_RegisterSong(void *data)
{
	HRESULT hr;
	MUS2MIDIState m2m;
	BYTE *pbTrackLength;
	DWORD dwTrackLength;
	IDirectMusicSegment *pSeg = NULL;
	DMUS_OBJECTDESC dmod;
	int i;

	if(!g_fPlayMusic || !data)
		return 0;

	ZeroMemory(&m2m, sizeof(MUS2MIDIState));

	// First, setup the MUS info...
	//
	memcpy(&m2m.m_mf, data, 16);
	m2m.m_mf.m_rgbMusicData = (byte *)_alloca(m2m.m_mf.m_wcbMusicData);
	memcpy(m2m.m_mf.m_rgbMusicData, ((byte *)data) + m2m.m_mf.m_wcbHeaderData, m2m.m_mf.m_wcbMusicData);
	m2m.m_pCurMUSEvent = m2m.m_mf.m_rgbMusicData;

	// Now, get the MIDI buffer ready...
	//
	m2m.m_dwcMIDIBufferLength = (m2m.m_mf.m_wcbMusicData * 2) + 32;
	m2m.m_rgmidEvents = (BYTE *)malloc(m2m.m_dwcMIDIBufferLength);

	// Add the MIDI header...
	//
	memcpy(m2m.m_rgmidEvents, g_MIDIHeader, sizeof(g_MIDIHeader));
	pbTrackLength = m2m.m_rgmidEvents + 18;
	m2m.m_pCurMIDEvent = m2m.m_rgmidEvents + sizeof(g_MIDIHeader);

	// Okay, add the MIDI events from the MUS file...
	//
	ConvertMUSEventsToMIDIEvents(&m2m);

	// Finally, add a MIDI "End-of-Track" event...
	// EOT: 0x00 0xFF 0x2F 0x00
	//
	m2m.m_pCurMIDEvent[0] = 0;
	m2m.m_pCurMIDEvent[1] = 0xFF;
	m2m.m_pCurMIDEvent[2] = 0x2F;
	m2m.m_pCurMIDEvent[3] = 0x00;

	m2m.m_pCurMIDEvent += 4;

	// Now, fill in the Track Length (which we left as all 0's before)
	// now that we know what the track length actually is...
	//
	dwTrackLength = m2m.m_pCurMIDEvent - (pbTrackLength + 4);
	pbTrackLength[0] = (BYTE)((dwTrackLength >> 24) & 0xFF);
	pbTrackLength[1] = (BYTE)((dwTrackLength >> 16) & 0xFF);
	pbTrackLength[2] = (BYTE)((dwTrackLength >> 8) & 0xFF);
	pbTrackLength[3] = (BYTE)(dwTrackLength & 0xFF);

	// That's it! Our MIDI buffer is ready!
	//
	dmod.dwSize = sizeof(DMUS_OBJECTDESC);
	dmod.dwValidData = DMUS_OBJ_CLASS | DMUS_OBJ_MEMORY;
	dmod.guidClass = CLSID_DirectMusicSegment;
	dmod.llMemLength = m2m.m_pCurMIDEvent - m2m.m_rgmidEvents;
	dmod.pbMemData = m2m.m_rgmidEvents;

	if(FAILED(hr = g_pDML->lpVtbl->GetObject(g_pDML, &dmod, &IID_IDirectMusicSegment, (void **)&pSeg)))
	{
		goto regErr;
	}

	if(FAILED(hr = pSeg->lpVtbl->SetParam(pSeg, &GUID_StandardMIDIFile, -1, 0, 0, (void *)g_pDMP)))
	{
		goto regErr;
	}

	if(FAILED(hr = pSeg->lpVtbl->SetParam(pSeg, &GUID_Download, -1, 0, 0, (void *)g_pDMP)))
	{
		goto regErr;
	}

	// Now find a slot in our SegmentToPointerMap array to store this Segment pointer
	// along with the memory I allocated for it...
	//
	for(i = 0; i < 32; i++)
	{
		if(g_rgS2P[i].m_pDMSeg == NULL)
		{
			g_rgS2P[i].m_pDMSeg = pSeg;
			g_rgS2P[i].m_rgbMIDIData = m2m.m_rgmidEvents;
			break;
		}
	}

	assert(i != 32);

	return (int)pSeg;

regErr:
	if(pSeg)
	{
		pSeg->lpVtbl->Release(pSeg);
	}
	if(m2m.m_rgmidEvents)
	{
		free(m2m.m_rgmidEvents);
	}

	return 0;

/*
  int rc = MUS_RegisterSong(data);
#ifdef SNDDEBUG
  if (rc<0) ST_Message("    MUS_Reg() returned %d\n", rc);
#endif
  return rc;
*/
}

void I_UnRegisterSong(int handle)
{
	IDirectMusicSegment *pSeg;
	int i;

	if(!g_fPlayMusic || !handle)
	{
		return;
	}

	pSeg = (IDirectMusicSegment *)handle;

	if(I_QrySongPlaying(handle))
	{
		OutputDebugString("!!!! WARNING: I_UnRegisterSong() called without a prior call to I_StopSong() !!!!\n");
		g_pDMP->lpVtbl->Stop(g_pDMP, pSeg, NULL, 0, 0);
	}

	for(i = 0; i < 32; i++)
	{
		if(g_rgS2P[i].m_pDMSeg == pSeg)
		{
			free(g_rgS2P[i].m_rgbMIDIData);
			g_rgS2P[i].m_pDMSeg = NULL;
			g_rgS2P[i].m_rgbMIDIData = NULL;
			break;
		}
	}

	pSeg->lpVtbl->Release(pSeg);

/*
	int rc = MUS_UnregisterSong(handle);
#ifdef SNDDEBUG
  if (rc < 0) ST_Message("    MUS_Unreg() returned %d\n", rc);
#endif
*/
}


// Stops a song.  MUST be called before I_UnregisterSong().

void I_StopSong(int handle)
{
	IDirectMusicSegment *pSeg;
	HRESULT hr;

	if(!g_fPlayMusic || !handle || !g_pDMP)
	{
		return;
	}

	pSeg = (IDirectMusicSegment *)handle;

	hr = g_pDMP->lpVtbl->Stop(g_pDMP, pSeg, NULL, 0, 0);
	while(I_QrySongPlaying(handle));

/*
	int rc;
  rc = MUS_StopSong(handle);
#ifdef SNDDEBUG
  if (rc < 0) ST_Message("    MUS_StopSong() returned %d\n", rc);
#endif
*/
}

void I_PlaySong(int handle, boolean looping)
{
	IDirectMusicSegment *pSeg;
	HRESULT hr;

	if(!g_fPlayMusic || !handle || !g_pDMP)
	{
		return;
	}

	pSeg = (IDirectMusicSegment *)handle;

	hr = pSeg->lpVtbl->SetStartPoint(pSeg, 0);
	hr = pSeg->lpVtbl->SetRepeats(pSeg, looping ? 999 : 1);

	hr = g_pDMP->lpVtbl->Stop(g_pDMP, NULL, NULL, 0, 0);
	hr = g_pDMP->lpVtbl->PlaySegment(g_pDMP, pSeg, 0, 0, NULL);

//	while(!I_QrySongPlaying(handle));
/*
	int rc;
  rc = MUS_ChainSong(handle, looping ? handle : -1);
#ifdef SNDDEBUG
  if (rc < 0) ST_Message("    MUS_ChainSong() returned %d\n", rc);
#endif
  rc = MUS_PlaySong(handle, snd_MusicVolume);
#ifdef SNDDEBUG
  if (rc < 0) ST_Message("    MUS_PlaySong() returned %d\n", rc);
#endif
*/
}




static DWORD UnpackVariableTimeLength(BYTE **ppBytes)
{
	DWORD dwLength = 0;
	BYTE bCur;
	int i = 0;

	do
	{
		bCur = **ppBytes;
		dwLength <<= 7;
		dwLength |= (bCur & 0x7F);
		i++;
		*ppBytes += 1;
	} while(bCur & 0x80);

	// No length should be more than 4 bytes total. If it is, we must
	// have &*@#!'ed-up somewhere...
	//
	assert(i <= 4);

	return dwLength;
}

static void PackVariableTimeLength(DWORD dwLen, BYTE **ppBytes)
{
	DWORD dwBuffer = dwLen & 0x7F;

	while( (dwLen >>= 7) )
	{
		dwBuffer <<= 8;
		dwBuffer |= (0x80 | (dwLen & 0x7F));
	}

	do
	{
		**ppBytes = (BYTE)(dwBuffer & 0xFF);
		*ppBytes += 1;
		if(dwBuffer & 0x80)
		{
			dwBuffer >>= 8;
		}
		else
		{
			break;
		}
	} while(1);
}


void WriteShortMIDIMessage(BYTE **ppCurMIDEvent, LONG lWait, BYTE bStatus, BYTE bData1, BYTE bData2)
{
	static BYTE bLastStatus = 0;

	// We assume Pack...() below will advance our pointer
	// the appropriate amount...
	//
	PackVariableTimeLength(lWait, ppCurMIDEvent);

	if(bStatus != bLastStatus)
	{
		**ppCurMIDEvent = bStatus;
		*ppCurMIDEvent += 1;
		bLastStatus = bStatus;
	}

	// Since data bytes aren't allowed to have the high-bit set, we'll use that
	// high-bit as an indication as to which data bytes to use...
	//
	// Don't use the data byte if the high bit is set.
	//

	if((bData1 & 0x80) == 0)
	{
		**ppCurMIDEvent = bData1;
		*ppCurMIDEvent += 1;
	}

	if((bData2 & 0x80) == 0)
	{
		**ppCurMIDEvent = bData2;
		*ppCurMIDEvent += 1;
	}
}


void HandleKeyUp(MUSEvent *pevt, MUS2MIDIState *pM2M)
{
	BYTE bChannel = pevt->m_event.m_bChannel;

#ifdef _DEBUG
		sprintf(pM2M->m_szCurrentOperation + strlen(pM2M->m_szCurrentOperation), "Key %d UP\n", (DWORD)(*(pM2M->m_pCurMUSEvent)));
#endif

	WriteShortMIDIMessage(&pM2M->m_pCurMIDEvent, 
						  pM2M->m_lNextWaitTime, 
						  (BYTE)(0x80 | bChannel),
						  *(pM2M->m_pCurMUSEvent),
						  pM2M->m_rgbChannelVolume[bChannel]);

	pM2M->m_pCurMUSEvent += 1;
}


void HandleKeyDown(MUSEvent *pevt, MUS2MIDIState *pM2M)
{
	BYTE bChannel = pevt->m_event.m_bChannel;
	BYTE bKey;
	BYTE bVolume;

	if(*(pM2M->m_pCurMUSEvent) & 0x80)
	{
		// The volume is specified, and will be the next byte.
		// We'll also store this volume as the relevent channel's
		// current volume.
		//
		bKey = *(pM2M->m_pCurMUSEvent) & 0x7F;
		bVolume = *(pM2M->m_pCurMUSEvent + 1);
		pM2M->m_pCurMUSEvent += 2;
		pM2M->m_rgbChannelVolume[bChannel] = bVolume;
#ifdef _DEBUG
		sprintf(pM2M->m_szCurrentOperation + strlen(pM2M->m_szCurrentOperation), "Key %d DOWN with VOLUME %d\n", (DWORD)bKey, (DWORD)bVolume);
#endif
	}
	else
	{
		// Volume is not specified. Just pull the current volume
		// out of the Channel's current data buffer.
		//
		bKey = *(pM2M->m_pCurMUSEvent);
		bVolume = pM2M->m_rgbChannelVolume[bChannel];
		pM2M->m_pCurMUSEvent += 1;
#ifdef _DEBUG
		sprintf(pM2M->m_szCurrentOperation + strlen(pM2M->m_szCurrentOperation), "Key %d DOWN\n", (DWORD)bKey);
#endif
	}

	WriteShortMIDIMessage(&pM2M->m_pCurMIDEvent, 
						  pM2M->m_lNextWaitTime, 
						  (BYTE)(0x90 | bChannel),
						  bKey,
						  bVolume);
}


void HandlePitchBend(MUSEvent *pevt, MUS2MIDIState *pM2M)
{
	WORD wPitchBend = *(pM2M->m_pCurMUSEvent);

	// In MUS, pitch bend values are 8-bit (0 to 0xFF), with 0x80 being "normal".
	// In MIDI, they are 14-bit (0 to 0x3FFF), with 0x2000 being "normal".
	//
	wPitchBend <<= 6;

#ifdef _DEBUG
		sprintf(pM2M->m_szCurrentOperation + strlen(pM2M->m_szCurrentOperation), "Pitch-Bend to %d\n", (DWORD)wPitchBend);
#endif

	WriteShortMIDIMessage(&pM2M->m_pCurMIDEvent,
						  pM2M->m_lNextWaitTime,
						  (BYTE)(0xE0 | pevt->m_event.m_bChannel),
						  (BYTE)(wPitchBend & 0x7F),
						  (BYTE)(wPitchBend >> 7));

	pM2M->m_pCurMUSEvent += 1;
}


void HandleSystemEvent(MUSEvent *pevt, MUS2MIDIState *pM2M)
{
	BYTE bData1;

	switch(*(pM2M->m_pCurMUSEvent))
	{
		case 10:
			bData1 = 0x78;
			break;

		case 11:
			bData1 = 0x7B;
			break;

		case 12:
			bData1 = 0x7E;
			break;

		case 13:
			bData1 = 0x7F;
			break;

		case 14:
			bData1 = 0x79;
			break;

		default:
			assert(false);
			break;
	}
#ifdef _DEBUG
		sprintf(pM2M->m_szCurrentOperation + strlen(pM2M->m_szCurrentOperation), "Valueless controller %d\n", (DWORD)bData1);
#endif

	WriteShortMIDIMessage(&pM2M->m_pCurMIDEvent,
						  pM2M->m_lNextWaitTime,
						  (BYTE)(0xB0 | pevt->m_event.m_bChannel),
						  bData1,
						  (BYTE)0x80);

	pM2M->m_pCurMUSEvent += 1;
}	


void HandleControllerEvent(MUSEvent *pevt, MUS2MIDIState *pM2M)
{
	BYTE bStatus;
	BYTE bData1, bData2;

	if(*(pM2M->m_pCurMUSEvent) == 0)
	{
		bStatus = 0xC0 | pevt->m_event.m_bChannel;
		bData1 = *(pM2M->m_pCurMUSEvent + 1);
		bData2 = 0x80;
#ifdef _DEBUG
		sprintf(pM2M->m_szCurrentOperation + strlen(pM2M->m_szCurrentOperation), "Patch Change to %d\n", (DWORD)bData1);
#endif
	}
	else
	{
		bStatus = 0xB0 | pevt->m_event.m_bChannel;
		bData2 = *(pM2M->m_pCurMUSEvent + 1);

		switch(*(pM2M->m_pCurMUSEvent))
		{
			case 1:
				bData1 = 0x00;
				break;

			case 2:
				bData1 = 0x01;
				break;

			case 3:
				bData1 = 0x07;
				break;

			case 4:
				bData1 = 0x0A;
				break;

			case 5:
				bData1 = 0x0B;
				break;

			case 6:
				bData1 = 0x5B;
				break;

			case 7:
				bData1 = 0x5D;
				break;

			case 8:
				bData1 = 0x40;
				break;

			case 9:
				bData1 = 0x43;
				break;

			default:
				assert(false);
				break;
		}

#ifdef _DEBUG
		sprintf(pM2M->m_szCurrentOperation + strlen(pM2M->m_szCurrentOperation), "Controller %d set value %d\n", (DWORD)bData1, (DWORD)bData2);
#endif
	}

	WriteShortMIDIMessage(&pM2M->m_pCurMIDEvent,
						  pM2M->m_lNextWaitTime,
						  bStatus,
						  bData1,
						  bData2);

	pM2M->m_pCurMUSEvent += 2;
}


void ConvertMUSEventsToMIDIEvents(MUS2MIDIState *pM2M)
{
	while(1)
	{
		MUSEvent evt;

		evt.m_bEvent = *(pM2M->m_pCurMUSEvent);

		// Put MUS channel 15 (its Drum channel) onto
		// channel 9 (GM Drum channel).
		//
		if(evt.m_event.m_bChannel == 15)
		{
			evt.m_event.m_bChannel = 9;
		}
		else if(evt.m_event.m_bChannel == 9)
		{
			evt.m_event.m_bChannel = 15;
		}

#ifdef _DEBUG
		sprintf(pM2M->m_szCurrentOperation, "Channel %ld: ", (DWORD)evt.m_event.m_bChannel);
#endif

		// We advance the MUS pointer in preparation for the function call.
		//
		pM2M->m_pCurMUSEvent += 1;

		switch(evt.m_event.m_bEventType)
		{
			case 0:
				HandleKeyUp(&evt, pM2M);
				break;

			case 1:
				HandleKeyDown(&evt, pM2M);
				break;

			case 2:
				HandlePitchBend(&evt, pM2M);
				break;

			case 3:
				HandleSystemEvent(&evt, pM2M);
				break;

			case 4:
				HandleControllerEvent(&evt, pM2M);
				break;

			case 6:
				return;
				break;

			default:
				assert(false);
				break;
		}

/*
#ifdef _DEBUG
		if(evt.m_event.m_bChannel == 15)
		{
			OutputDebugString(pM2M->m_szCurrentOperation);
		}
#endif
*/
		// Verify we haven't written past our MIDIEVENT array...
		//
		assert(pM2M->m_pCurMIDEvent < pM2M->m_rgmidEvents + pM2M->m_dwcMIDIBufferLength);

		// We expect that the function call above advanced the MUS
		// pointer as necessary.
		//
		if(evt.m_event.m_bLast)
		{
			pM2M->m_lNextWaitTime = UnpackVariableTimeLength(&pM2M->m_pCurMUSEvent);
		}
		else
		{
			pM2M->m_lNextWaitTime = 0;
		}
	};
}
