
//**************************************************************************
//**
//** I_IBM.C
//**
//**************************************************************************

#include <dos.h>
#include <conio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <ddraw.h>
#include <dinput.h>
#include <dsound.h>
#include <d3dtypes.h>
#include <process.h>
#include <math.h>
#include <assert.h>
#define __BYTEBOOL__
#include <setjmp.h>
#include <crtdbg.h>
#include "h2def.h"
#include "r_local.h"
#include "p_local.h"    // for P_AproxDistance
#include "sounds.h"
#include "soundst.h"
#include "SFXInfo.h"
#include "i_sound.h"
#include "st_start.h"


HRESULT I_CreateDSBuffer(int id);
void I_DestroyDSBuffer(int id);
void I_UpdateSoundInfo(channel_t *channel);
void I_UpdateListenerInfoAndCommit(mobj_t *listener);
static HRESULT WINAPI EnumModesCallback2(DDSURFACEDESC2 *pDDSD, void *pvData);

// Macros

#define DEFAULT_ARCHIVEPATH     "o:\\sound\\archive\\"
#define PRIORITY_MAX_ADJUST 10
#define DIST_ADJUST (MAX_SND_DIST/PRIORITY_MAX_ADJUST)

//#define SEQ_ADDR 0x3C4
//#define SEQ_DATA 0x3C5
//#define REG_MAPMASK 0x02

//#define MASK_PLANE0 0x01
//#define MASK_PLANE1 0x02
//#define MASK_PLANE2 0x04
//#define MASK_PLANE3 0x08

#define P0OFFSET 38400*0
#define P1OFFSET 38400*1
#define P2OFFSET 38400*2
#define P3OFFSET 38400*3

//#define VID_INT 0x10
//#define VB_SYNC {while(!(inp(0x3da)&8)); while(inp(0x3da)&8);}
//#define BITPLANE(p) (outp(SEQ_ADDR,REG_MAPMASK),outp(SEQ_DATA,(p)))

#define MAX_ARGS 64

//#define NOKBD
//#define NOTIMER

// Private Data

static IDirectDrawSurface4 *g_pDDS;
static IDirectDrawPalette *g_pDDP;
static HINSTANCE g_hInst;
static const char gc_szTTTEName[] = "TimerThreadTerminationEvent";
static const char gc_szInputThreadReady[] = "InputThreadReadyEvent";
static const char gc_szInputThreadTerminate[] = "InputThreadTerminateEvent";
static const char gc_szDIKeyboardEvent[] = "DirectInputKeyboardEvent";
static const char gc_szDIMouseEvent[] = "DirectInputMouseEvent";
static HANDLE g_hTTTE;
static HANDLE g_hTimerThread;
static PALETTEENTRY g_palLastSet[256];
static PALETTEENTRY g_palOriginal[256];
static HANDLE g_hInputThreadReady;
static HANDLE g_hInputThreadTerminate;
static HANDLE g_hInputThread;

// Public Data

int DisplayTicker = 0;
IDirectDrawSurface4 *g_pDDSScreen = NULL;	// DirectDrawSurface which wraps 'screen'
IDirectDraw4 *g_pDD;
//IDirectDrawClipper *g_pDDC = NULL;
HWND g_hWnd;
DWORD *g_rgScreenWidth = NULL;
DWORD *g_rgScreenHeight = NULL;
DWORD g_dwcResolutions = 0;

int g_ScreenWidth = 640;
int g_ScreenHeight = 480;

boolean g_fPlayInWindow = false;
volatile boolean g_fActive = true;

extern boolean fShowFPS;

// Code

void I_StartupMouseAndKeyboard(void);


LRESULT CALLBACK HexenWindowProc(HWND hwnd, UINT msg, WPARAM wParam, LPARAM lParam)
{
	switch(msg)
	{
		case WM_CLOSE:
			g_fActive = true;
			I_Quit();
			break;

		case WM_DESTROY:
			g_fActive = true;
			break;

		case WM_MOUSEMOVE:
			// This is just so you don't get a big ugly cursor on top of your
			// game window. The actual retrieval of mouse information is via
			// DirectInput and is read from a separate thread.
			//
			SetCursor(NULL);
			break;

		case WM_ACTIVATE:
			g_fActive = LOWORD(wParam) == WA_INACTIVE ? false : true;
			break;

		case WM_ACTIVATEAPP:
			g_fActive = (BOOL)wParam;
			break;

/*
		case WM_PALETTECHANGED:
			if ((HWND) wParam == hwnd)
				break;
			// Fall through to WM_QUERYNEWPALETTE

		case WM_QUERYNEWPALETTE:
			// Install our palette here
			if(g_pDDP && g_pDDS)
			{
				IDirectDrawSurface4_SetPalette(g_pDDS, g_pDDP);
			}
			break;
*/
	}

	return DefWindowProc(hwnd, msg, wParam, lParam);
}


int WINAPI WinMain(HINSTANCE hInst, HINSTANCE hPrevInst, LPSTR szUselessCommandLine, int nCmdShow)
{
	char *szToken;
	char *szCmdLine = GetCommandLine();
	int w, h;
	WNDCLASS wndcls;

	g_hInst = hInst;
	g_pDD = NULL;
	g_pDDS = NULL;
	g_pDDP = NULL;
	myargv = (char **)malloc(MAX_ARGS * sizeof(char *));
	myargc = 0;

	if(*szCmdLine == '"')
	{
		// Strip the Quote characters out of argv[0]
		// in szCmdLine...
		//
		szCmdLine++;
		szToken = strchr(szCmdLine, '"');
		if(!szToken)
		{
			free(myargv);
			return -1;
		}
		MoveMemory(szToken, szToken + 1, strlen(szToken + 1) + 1);
	}

	// Note: Assignment is intentional...
	//
	if(szToken = strrchr(szCmdLine, '\\'))
	{
		// Path is present -- strip it!
		//
		szCmdLine = szToken + 1;
	}

	szToken = strtok(szCmdLine, " ");
	while(szToken != NULL)
	{
		if(myargc >= MAX_ARGS)
		{
			free(myargv);
			return -1;
		}
		myargv[myargc++] = strdup(szToken);
		szToken = strtok(NULL, " ");
	}

	// Ok, we've got the command-line arguments setup like the old DOS days...
	//
	// Parse out my own params that are not in the original Hexen...
	//
	w = M_CheckParm("-width");
	h = M_CheckParm("-height");

	if(w && h && (w + 1) < myargc && (h + 1) < myargc)
	{
		g_ScreenWidth = atoi(myargv[w + 1]);
		g_ScreenHeight = atoi(myargv[h + 1]);
	}

	if(g_ScreenWidth & 0x0F)
	{
		MessageBox(NULL, "The screen width must be a multiple of 16.", "Bad Screen Resolution", MB_OK | MB_ICONEXCLAMATION);
		free(myargv);
		return -1;
	}

	if(g_ScreenWidth > 1600 || g_ScreenHeight > 1200)
	{
		MessageBox(NULL, "The maximum screen dimensions are 1600x1200.", "Bad Screen Resolution", MB_OK | MB_ICONEXCLAMATION);
		free(myargv);
		return -1;
	}

	if(g_ScreenWidth < 320 || g_ScreenHeight < 200)
	{
		MessageBox(NULL, "The minimum screen dimensions are 320x200.", "Bad Screen Resolution", MB_OK | MB_ICONEXCLAMATION);
		free(myargv);
		return -1;
	}

	// Get ready for the Window...
	//

	ZeroMemory(&wndcls, sizeof(WNDCLASS));

	wndcls.hInstance = hInst;
	wndcls.style = CS_HREDRAW | CS_VREDRAW;
	wndcls.lpfnWndProc = HexenWindowProc;
	wndcls.lpszClassName = "HexenWindowClass";
	wndcls.hbrBackground = (HBRUSH)GetStockObject(BLACK_BRUSH);
	wndcls.hIcon = LoadIcon(NULL, IDI_APPLICATION);

	if(RegisterClass(&wndcls) == 0)
	{
		free(myargv);
		return -1;
	}

	H2_Main();

	// EVERYTHING should be freed by the time we get here...
	// (i.e. Window closed, DirectDraw released, etc.)
	//

	_ASSERTE(g_pDD == NULL);
	_ASSERTE(g_pDDS == NULL);
	_ASSERTE(g_pDDP == NULL);
	_ASSERTE(g_hTTTE == NULL);
	_ASSERTE(g_hTimerThread == NULL);
	_ASSERTE(g_hInputThreadReady == NULL);
	_ASSERTE(g_hInputThreadTerminate == NULL);
	_ASSERTE(g_hInputThread == NULL);

	for( ; myargc > 0; myargc--)
	{
		free(myargv[myargc - 1]);
	}
	free(myargv);

	return 0;
}

/*
void main(int argc, char **argv)
{
	myargc = argc;
	myargv = argv;
	H2_Main();
}
*/

void I_StartupNet (void);
void I_ShutdownNet (void);
void I_ReadExternDriver(void);

//void I_ReadMouse (void);

extern  int     usemouse, usejoystick;

extern void **lumpcache;

int oldTic;

DWORD I_GetWindowsTime(void)
{
	return timeGetTime();
}


/*
===============================================================================

		MUSIC & SFX API

===============================================================================
*/

//static channel_t channel[MAX_CHANNELS];

//static int rs; //the current registered song.
//int mus_song = -1;
//int mus_lumpnum;
//void *mus_sndptr;
//byte *soundCurve;

extern sfxinfo_t S_sfx[];
extern musicinfo_t S_music[];

extern int snd_MusicDevice;
extern int snd_SfxDevice;
extern int snd_MaxVolume;
extern int snd_MusicVolume;
extern int snd_Channels;

extern int startepisode;
extern int startmap;

static channel_t Channel[MAX_CHANNELS];
static int RegisteredSong; //the current registered song.
static int NextCleanup;
static boolean MusicPaused;
static int Mus_Song = -1;
static int Mus_LumpNum;
static void *Mus_SndPtr;

static boolean UseSndScript;
static char ArchivePath[128];

byte *SoundCurve;


// int AmbChan;

//==========================================================================
//
// S_Start
//
//==========================================================================

void S_Start(void)
{
	S_StopAllSound();
	S_StartSong(gamemap, true);
}

//==========================================================================
//
// S_StartSong
//
//==========================================================================

void S_StartSong(int song, boolean loop)
{
	char *songLump;

	if(song == Mus_Song)
	{ // don't replay an old song
		return;
	}
	if(RegisteredSong)
	{
		I_StopSong(RegisteredSong);
		I_UnRegisterSong(RegisteredSong);
		if(UseSndScript)
		{
			Z_Free(Mus_SndPtr);
		}
		else
		{
			Z_ChangeTag(lumpcache[Mus_LumpNum], PU_CACHE);
		}
		RegisteredSong = 0;
	}
	songLump = P_GetMapSongLump(song);
	if(!songLump)
	{
		return;
	}
	if(UseSndScript)
	{
		char name[128];
		sprintf(name, "%s%s.lmp", ArchivePath, songLump);
		M_ReadFile(name, (byte **)&Mus_SndPtr);
		Mus_LumpNum = W_GetNumForName(songLump);
	}
	else
	{
		Mus_LumpNum = W_GetNumForName(songLump);
		Mus_SndPtr = W_CacheLumpNum(Mus_LumpNum, PU_MUSIC);
	}
	RegisteredSong = I_RegisterSong(Mus_SndPtr, Mus_LumpNum);
	I_PlaySong(RegisteredSong, loop); // 'true' denotes endless looping.
	Mus_Song = song;
}

//==========================================================================
//
// S_StartSongName
//
//==========================================================================

void S_StartSongName(char *songLump, boolean loop)
{
	if(!songLump)
	{
		return;
	}
	if(RegisteredSong)
	{
		I_StopSong(RegisteredSong);
		I_UnRegisterSong(RegisteredSong);
		if(UseSndScript)
		{
			Z_Free(Mus_SndPtr);
		}
		else
		{
			Z_ChangeTag(lumpcache[Mus_LumpNum], PU_CACHE);
		}
		RegisteredSong = 0;
	}
	if(UseSndScript)
	{
		char name[128];
		sprintf(name, "%s%s.lmp", ArchivePath, songLump);
		M_ReadFile(name, (byte **)&Mus_SndPtr);
		Mus_LumpNum = W_GetNumForName(songLump);
	}
	else
	{
		Mus_LumpNum = W_GetNumForName(songLump);
		Mus_SndPtr = W_CacheLumpNum(Mus_LumpNum, PU_MUSIC);
	}
	RegisteredSong = I_RegisterSong(Mus_SndPtr, Mus_LumpNum);
	I_PlaySong(RegisteredSong, loop); // 'true' denotes endless looping.
	Mus_Song = -1;
}

//==========================================================================
//
// S_GetSoundID
//
//==========================================================================

int S_GetSoundID(char *name)
{
	int i;

	for(i = 0; i < NUMSFX; i++)
	{
		if(!strcmp(S_sfx[i].tagName, name))
		{
			return i;
		}
	}
	return 0;
}

//==========================================================================
//
// S_StartSound
//
//==========================================================================

void S_StartSound(mobj_t *origin, int sound_id)
{
	S_StartSoundAtVolume(origin, sound_id, 127);
}

//==========================================================================
//
// S_StopSoundID
//
//==========================================================================

boolean S_StopSoundID(int sound_id, int priority)
{
	int i;
	int lp; //least priority
	int found;

	if(S_sfx[sound_id].numchannels == -1)
	{
		return(true);
	}
	lp = -1; //denote the argument sound_id
	found = 0;
	for(i=0; i<snd_Channels; i++)
	{
		if(Channel[i].sound_id == sound_id && Channel[i].mo)
		{
			found++; //found one.  Now, should we replace it??
			if(priority >= Channel[i].priority)
			{ // if we're gonna kill one, then this'll be it
				lp = i;
				priority = Channel[i].priority;
			}
		}
	}
	if(found < S_sfx[sound_id].numchannels)
	{
		return(true);
	}
	else if(lp == -1)
	{
		return(false); // don't replace any sounds
	}
	if(Channel[lp].handle)
	{
		if(I_SoundIsPlaying(Channel[lp].handle))
		{
			I_StopSound(Channel[lp].handle);
		}
		if(S_sfx[Channel[lp].sound_id].usefulness > 0)
		{
			S_sfx[Channel[lp].sound_id].usefulness--;
		}
		Channel[lp].mo = NULL;
	}

	return(true);
}

//==========================================================================
//
// S_StartSoundAtVolume
//
//==========================================================================

void S_StartSoundAtVolume(mobj_t *origin, int sound_id, int volume)
{
	int dist;
//	int vol;
	int i;
	int priority;
//	int sep;
//	int angle;
	int absx;
	int absy;
	boolean fOriginIsPlayer;

	static int sndcount = 0;
	int chan;

	if(sound_id == 0 || snd_MaxVolume == 0 || volume == 0)
		return;

	if(origin == NULL || origin == players[displayplayer].mo)
	{
		origin = players[displayplayer].mo;
		dist = 0;
		fOriginIsPlayer = true;
	}
	else
	{
		// calculate the distance before other stuff so that we can throw out
		// sounds that are beyond the hearing range.
		//
		absx = abs(origin->x-players[displayplayer].mo->x);
		absy = abs(origin->y-players[displayplayer].mo->y);
		dist = absx+absy-(absx > absy ? absy>>1 : absx>>1);
		dist >>= FRACBITS;
		fOriginIsPlayer = false;
	}

	if(dist >= MAX_SND_DIST)
	{
	  return; // sound is beyond the hearing range...
	}
	if(dist < 0)
	{
		dist = 0;
	}
	priority = S_sfx[sound_id].priority;
	priority *= (PRIORITY_MAX_ADJUST-(dist/DIST_ADJUST));
	if(!S_StopSoundID(sound_id, priority))
	{
		return; // other sounds have greater priority
	}
	if(origin)
	{
		for(i=0; i<snd_Channels; i++)
		{
			if(origin->player)
			{
				i = snd_Channels;
				break; // let the player have more than one sound.
			}
			if(origin == Channel[i].mo)
			{ // only allow other mobjs one sound
				S_StopSound(Channel[i].mo);
				break;
			}
		}
	}
	else
	{
		i = snd_Channels;
	}
	if(i >= snd_Channels)
	{
		for(i = 0; i < snd_Channels; i++)
		{
			if(Channel[i].mo == NULL)
			{
				break;
			}
		}
		if(i >= snd_Channels)
		{
			// look for a lower priority sound to replace.
			sndcount++;
			if(sndcount >= snd_Channels)
			{
				sndcount = 0;
			}
			for(chan = 0; chan < snd_Channels; chan++)
			{
				i = (sndcount+chan)%snd_Channels;
				if(priority >= Channel[i].priority)
				{
					chan = -1; //denote that sound should be replaced.
					break;
				}
			}
			if(chan != -1)
			{
				return; //no free channels.
			}
			else //replace the lower priority sound.
			{
				if(Channel[i].handle)
				{
					if(I_SoundIsPlaying(Channel[i].handle))
					{
						I_StopSound(Channel[i].handle);
					}
					if(S_sfx[Channel[i].sound_id].usefulness > 0)
					{
						S_sfx[Channel[i].sound_id].usefulness--;
					}
				}
			}
		}
	}
	if(S_sfx[sound_id].lumpnum == 0)
	{
		S_sfx[sound_id].lumpnum = I_GetSfxLumpNum(&S_sfx[sound_id]);
	}
	if(S_sfx[sound_id].snd_ptr == NULL)
	{
		HRESULT hr;

		if(UseSndScript)
		{
			char name[128];
			HRESULT hr;

			sprintf(name, "%s%s.lmp", ArchivePath, S_sfx[sound_id].lumpname);
			M_ReadFile(name, (byte **)&S_sfx[sound_id].snd_ptr);
			if(FAILED(hr = I_CreateDSBuffer(sound_id)))
			{
				I_Error("Could not create sound buffer for sound! (Error code 0x%08x)", hr);
			}
		}
		else
		{
			S_sfx[sound_id].snd_ptr = W_CacheLumpNum(S_sfx[sound_id].lumpnum,
				PU_SOUND);
			if(FAILED(hr = I_CreateDSBuffer(sound_id)))
			{
				I_Error("Could not create sound buffer for sound! (Error code 0x%08x)", hr);
			}
		}
/*
#ifdef __WATCOMC__
		_dpmi_lockregion(S_sfx[sound_id].snd_ptr,
			lumpinfo[S_sfx[sound_id].lumpnum].size);
#endif
*/
	}

/*
	vol = (SoundCurve[dist]*(snd_MaxVolume*8)*volume)>>14;
	if(origin == players[displayplayer].mo)
	{
		sep = 128;
//              vol = (volume*(snd_MaxVolume+1)*8)>>7;
	}
	else
	{
		angle = R_PointToAngle2(players[displayplayer].mo->x,
			players[displayplayer].mo->y, Channel[i].mo->x, Channel[i].mo->y);
		angle = (angle-viewangle)>>24;
		sep = angle*2-128;
		if(sep < 64)
			sep = -sep;
		if(sep > 192)
			sep = 512-sep;
//              vol = SoundCurve[dist];
	}

	if(S_sfx[sound_id].changePitch)
	{
		Channel[i].pitch = (byte)(127+(M_Random()&7)-(M_Random()&7));
	}
	else
	{
		Channel[i].pitch = 127;
	}
*/

	Channel[i].mo = origin;
	Channel[i].sound_id = sound_id;
	Channel[i].priority = priority;
	Channel[i].volume = volume;

	if(S_sfx[sound_id].changePitch)
	{
		Channel[i].pitch = (SND_SAMPLERATE + 
						   ((M_Random() & 7) * 64) - 
						   ((M_Random() & 7) * 64));
	}
	else
	{
		Channel[i].pitch = SND_SAMPLERATE;
	}

	I_StartSound(sound_id, &Channel[i], fOriginIsPlayer);

	if(S_sfx[sound_id].usefulness < 0)
	{
		S_sfx[sound_id].usefulness = 1;
	}
	else
	{
		S_sfx[sound_id].usefulness++;
	}
}


//==========================================================================
//
// S_StopSound
//
//==========================================================================

void S_StopSound(mobj_t *origin)
{
	int i;

	for(i=0;i<snd_Channels;i++)
	{
		if(Channel[i].mo == origin)
		{
			I_StopSound(Channel[i].handle);
			if(S_sfx[Channel[i].sound_id].usefulness > 0)
			{
				S_sfx[Channel[i].sound_id].usefulness--;
			}
			Channel[i].handle = 0;
			Channel[i].mo = NULL;
		}
	}
}

//==========================================================================
//
// S_StopAllSound
//
//==========================================================================

void S_StopAllSound(void)
{
	int i;

	//stop all sounds
	for(i=0; i < snd_Channels; i++)
	{
		if(Channel[i].handle)
		{
			S_StopSound(Channel[i].mo);
		}
	}
	memset(Channel, 0, 8*sizeof(channel_t));
}

//==========================================================================
//
// S_SoundLink
//
//==========================================================================

void S_SoundLink(mobj_t *oldactor, mobj_t *newactor)
{
	int i;

	for(i=0;i<snd_Channels;i++)
	{
		if(Channel[i].mo == oldactor)
			Channel[i].mo = newactor;
	}
}

//==========================================================================
//
// S_PauseSound
//
//==========================================================================

void S_PauseSound(void)
{
	I_PauseSong(RegisteredSong);
}

//==========================================================================
//
// S_ResumeSound
//
//==========================================================================

void S_ResumeSound(void)
{
	I_ResumeSong(RegisteredSong);
}

//==========================================================================
//
// S_UpdateSounds
//
//==========================================================================

void S_UpdateSounds(mobj_t *listener)
{
	int i;
	int dist;
//	int vol;
//	int angle;
//	int sep;
	int priority;
	int absx;
	int absy;

	if(snd_MaxVolume == 0)
	{
		return;
	}

	// Update any Sequences
	SN_UpdateActiveSequences();

	if(NextCleanup < gametic)
	{
		if(UseSndScript)
		{
			for(i = 0; i < NUMSFX; i++)
			{
				if(S_sfx[i].usefulness == 0 && S_sfx[i].snd_ptr)
				{
					S_sfx[i].usefulness = -1;
				}
			}
		}
		else
		{
			for(i = 0; i < NUMSFX; i++)
			{
				if(S_sfx[i].usefulness == 0 && S_sfx[i].snd_ptr)
				{
					if(lumpcache[S_sfx[i].lumpnum])
					{
						if(((memblock_t *)((byte*)
							(lumpcache[S_sfx[i].lumpnum])-
							sizeof(memblock_t)))->id == 0x1d4a11)
						{ // taken directly from the Z_ChangeTag macro
							Z_ChangeTag2(lumpcache[S_sfx[i].lumpnum],
								PU_CACHE);
						}
					}
					S_sfx[i].usefulness = -1;
					S_sfx[i].snd_ptr = NULL;
					I_DestroyDSBuffer(i);
				}
			}
		}
		NextCleanup = gametic+35*30; // every 30 seconds
	}

	for(i=0;i<snd_Channels;i++)
	{
		if(!Channel[i].handle || S_sfx[Channel[i].sound_id].usefulness == -1)
		{
			continue;
		}
		if(!I_SoundIsPlaying(Channel[i].handle))
		{
			if(S_sfx[Channel[i].sound_id].usefulness > 0)
			{
				S_sfx[Channel[i].sound_id].usefulness--;
			}
			Channel[i].handle = 0;
			Channel[i].mo = NULL;
			Channel[i].sound_id = 0;
		}
		if(Channel[i].mo == NULL || Channel[i].sound_id == 0
			|| Channel[i].mo == listener)
		{
			continue;
		}
		else
		{
			absx = abs(Channel[i].mo->x-listener->x);
			absy = abs(Channel[i].mo->y-listener->y);
			dist = absx+absy-(absx > absy ? absy>>1 : absx>>1);
			dist >>= FRACBITS;

			if(dist >= MAX_SND_DIST)
			{
				S_StopSound(Channel[i].mo);
				continue;
			}

			I_UpdateSoundInfo(&Channel[i]);

/*
			if(dist < 0)
			{
				dist = 0;
			}
			//vol = SoundCurve[dist];
			vol = (SoundCurve[dist]*(snd_MaxVolume*8)*Channel[i].volume)>>14;
			if(Channel[i].mo == listener)
			{
				sep = 128;
			}
			else
			{
				angle = R_PointToAngle2(listener->x, listener->y,
								Channel[i].mo->x, Channel[i].mo->y);
				angle = (angle-viewangle)>>24;
				sep = angle*2-128;
				if(sep < 64)
					sep = -sep;
				if(sep > 192)
					sep = 512-sep;
			}
			I_UpdateSoundParams(Channel[i].handle, vol, sep,
				Channel[i].pitch);
*/
			priority = S_sfx[Channel[i].sound_id].priority;
			priority *= PRIORITY_MAX_ADJUST-(dist/DIST_ADJUST);
			Channel[i].priority = priority;
		}
	}

	// This call will update the listener location, orientation, etc.,
	// and commit the changes to the other sounds (made above) as well...
	//
	I_UpdateListenerInfoAndCommit(listener);
}

//==========================================================================
//
// S_Init
//
//==========================================================================

void S_Init(void)
{
	SoundCurve = W_CacheLumpName("SNDCURVE", PU_STATIC);
//      SoundCurve = Z_Malloc(MAX_SND_DIST, PU_STATIC, NULL);

	ZeroMemory(Channel, sizeof(channel_t) * MAX_CHANNELS);

	I_StartupSound();

	// Go ahead and let them specify more than 8 channels...
	//
/*
	if(snd_Channels > 8)
	{
		snd_Channels = 8;
	}
*/

	I_SetChannels(snd_Channels);
	I_SetMusicVolume(snd_MusicVolume);

	// Attempt to setup CD music
/*
	if(snd_MusicDevice == snd_CDMUSIC)
	{
   		ST_Message("failed.\n");
	}
*/
}

//==========================================================================
//
// S_GetChannelInfo
//
//==========================================================================

void S_GetChannelInfo(SoundInfo_t *s)
{
	int i;
	ChanInfo_t *c;

	s->channelCount = snd_Channels;
	s->musicVolume = snd_MusicVolume;
	s->soundVolume = snd_MaxVolume;
	for(i = 0; i < snd_Channels; i++)
	{
		c = &s->chan[i];
		c->id = Channel[i].sound_id;
		c->priority = Channel[i].priority;
		c->name = S_sfx[c->id].lumpname;
		c->mo = Channel[i].mo;
		c->distance = P_AproxDistance(c->mo->x-viewx, c->mo->y-viewy)
			>>FRACBITS;
	}
}

//==========================================================================
//
// S_GetSoundPlayingInfo
//
//==========================================================================

boolean S_GetSoundPlayingInfo(mobj_t *mobj, int sound_id)
{
	int i;

	for(i = 0; i < snd_Channels; i++)
	{
		if(Channel[i].sound_id == sound_id && Channel[i].mo == mobj)
		{
			if(I_SoundIsPlaying(Channel[i].handle))
			{
				return true;
			}
		}
	}

	return false;
}

//==========================================================================
//
// S_SetMusicVolume
//
//==========================================================================

void S_SetMusicVolume(void)
{
	I_SetMusicVolume(snd_MusicVolume);

	if(snd_MusicVolume == 0)
	{
		I_PauseSong(RegisteredSong);
		MusicPaused = true;
	}
	else if(MusicPaused)
	{
		I_ResumeSong(RegisteredSong);
		MusicPaused = false;
	}
}

//==========================================================================
//
// S_ShutDown
//
//==========================================================================

void S_ShutDown(void)
{
	I_ShutdownSound();

/*
	extern int tsm_ID;
	if(tsm_ID != -1)
	{
		I_StopSong(RegisteredSong);
		I_UnRegisterSong(RegisteredSong);
		I_ShutdownSound();
	}
*/
}

//==========================================================================
//
// S_InitScript
//
//==========================================================================

void S_InitScript(void)
{
	int p;
	int i;

	strcpy(ArchivePath, DEFAULT_ARCHIVEPATH);
	if(!(p = M_CheckParm("-devsnd")))
	{
		UseSndScript = false;
		SC_OpenLump("sndinfo");
	}
	else
	{
		UseSndScript = true;
		SC_OpenFile(myargv[p+1]);
	}
	while(SC_GetString())
	{
		if(*sc_String == '$')
		{
			if(!stricmp(sc_String, "$ARCHIVEPATH"))
			{
				SC_MustGetString();
				strcpy(ArchivePath, sc_String);
			}
			else if(!stricmp(sc_String, "$MAP"))
			{
				SC_MustGetNumber();
				SC_MustGetString();
				if(sc_Number)
				{
					P_PutMapSongLump(sc_Number, sc_String);
				}
			}
			continue;
		}
		else
		{
			for(i = 0; i < NUMSFX; i++)
			{
				if(!strcmp(S_sfx[i].tagName, sc_String))
				{
					SC_MustGetString();
					if(*sc_String != '?')
					{
						strcpy(S_sfx[i].lumpname, sc_String);
					}
					else
					{
						strcpy(S_sfx[i].lumpname, "default");
					}
					break;
				}
			}
			if(i == NUMSFX)
			{
				SC_MustGetString();
			}
		}
	}
	SC_Close();

	for(i = 0; i < NUMSFX; i++)
	{
		if(!strcmp(S_sfx[i].lumpname, ""))
		{
			strcpy(S_sfx[i].lumpname, "default");
		}
	}
}


/*
============================================================================

							CONSTANTS

============================================================================
*/

#define SC_INDEX                0x3C4
#define SC_RESET                0
#define SC_CLOCK                1
#define SC_MAPMASK              2
#define SC_CHARMAP              3
#define SC_MEMMODE              4

#define CRTC_INDEX              0x3D4
#define CRTC_H_TOTAL    0
#define CRTC_H_DISPEND  1
#define CRTC_H_BLANK    2
#define CRTC_H_ENDBLANK 3
#define CRTC_H_RETRACE  4
#define CRTC_H_ENDRETRACE 5
#define CRTC_V_TOTAL    6
#define CRTC_OVERFLOW   7
#define CRTC_ROWSCAN    8
#define CRTC_MAXSCANLINE 9
#define CRTC_CURSORSTART 10
#define CRTC_CURSOREND  11
#define CRTC_STARTHIGH  12
#define CRTC_STARTLOW   13
#define CRTC_CURSORHIGH 14
#define CRTC_CURSORLOW  15
#define CRTC_V_RETRACE  16
#define CRTC_V_ENDRETRACE 17
#define CRTC_V_DISPEND  18
#define CRTC_OFFSET             19
#define CRTC_UNDERLINE  20
#define CRTC_V_BLANK    21
#define CRTC_V_ENDBLANK 22
#define CRTC_MODE               23
#define CRTC_LINECOMPARE 24


#define GC_INDEX                0x3CE
#define GC_SETRESET             0
#define GC_ENABLESETRESET 1
#define GC_COLORCOMPARE 2
#define GC_DATAROTATE   3
#define GC_READMAP              4
#define GC_MODE                 5
#define GC_MISCELLANEOUS 6
#define GC_COLORDONTCARE 7
#define GC_BITMASK              8

#define ATR_INDEX               0x3c0
#define ATR_MODE                16
#define ATR_OVERSCAN    17
#define ATR_COLORPLANEENABLE 18
#define ATR_PELPAN              19
#define ATR_COLORSELECT 20

#define STATUS_REGISTER_1    0x3da

#define PEL_WRITE_ADR   0x3c8
#define PEL_READ_ADR    0x3c7
#define PEL_DATA                0x3c9
#define PEL_MASK                0x3c6

// boolean grmode;

//==================================================

#define VBLCOUNTER              34000           // hardware tics to a frame


#define TIMERINT 8
#define KEYBOARDINT 9

#define CRTCOFF (_inbyte(STATUS_REGISTER_1)&1)

#define MOUSEB1 1
#define MOUSEB2 2
#define MOUSEB3 4

boolean mousepresent;
//static  int tsm_ID = -1; // tsm init flag

//===============================

volatile int ticcount;

boolean novideo; // if true, stay in text mode for debugging

//#define KBDQUESIZE 32
// Note that the keyboard queue is cyclical: it doesn't overflow, it just "wraps around"
// and trashes the oldest data if too much data is sent to it...
//
//volatile byte keyboardque[KBDQUESIZE];
//volatile int kbdtail, kbdhead;

#define KEY_LSHIFT      0xfe

#define KEY_INS         (0x80+0x52)
#define KEY_DEL         (0x80+0x53)
#define KEY_PGUP        (0x80+0x49)
#define KEY_PGDN        (0x80+0x51)
#define KEY_HOME        (0x80+0x47)
#define KEY_END         (0x80+0x4f)

#define SC_RSHIFT       0x36
#define SC_LSHIFT       0x2a

byte        scantokey[128] =
					{
//  0        1       2       3       4       5       6       7
//  8        9       A       B       C       D       E       F
	0  ,    27,     '1',    '2',    '3',    '4',    '5',    '6',
	'7',    '8',    '9',    '0',    '-',    '=',    KEY_BACKSPACE, 9, // 0
	'q',    'w',    'e',    'r',    't',    'y',    'u',    'i',
	'o',    'p',    '[',    ']',    13 ,    KEY_RCTRL,'a',  's',      // 1
	'd',    'f',    'g',    'h',    'j',    'k',    'l',    ';',
	39 ,    '`',    KEY_LSHIFT,92,  'z',    'x',    'c',    'v',      // 2
	'b',    'n',    'm',    ',',    '.',    '/',    KEY_RSHIFT,'*',
	KEY_RALT,' ',   0  ,    KEY_F1, KEY_F2, KEY_F3, KEY_F4, KEY_F5,   // 3
	KEY_F6, KEY_F7, KEY_F8, KEY_F9, KEY_F10,0  ,    0  , KEY_HOME,
	KEY_UPARROW,KEY_PGUP,'-',KEY_LEFTARROW,'5',KEY_RIGHTARROW,'+',KEY_END, //4
	KEY_DOWNARROW,KEY_PGDN,KEY_INS,KEY_DEL,0,0,             0,              KEY_F11,
	KEY_F12,0  ,    0  ,    0  ,    0  ,    0  ,    0  ,    0,        // 5
	0  ,    0  ,    0  ,    0  ,    0  ,    0  ,    0  ,    0,
	0  ,    0  ,    0  ,    0  ,    0  ,    0  ,    0  ,    0,        // 6
	0  ,    0  ,    0  ,    0  ,    0  ,    0  ,    0  ,    0,
	0  ,    0  ,    0  ,    0  ,    0  ,    0  ,    0  ,    0         // 7
					};

//==========================================================================

//--------------------------------------------------------------------------
//
// FUNC I_GetTime
//
// Returns time in 1/35th second tics.
//
//--------------------------------------------------------------------------

int I_GetTime (void)
{
#ifdef NOTIMER
	ticcount++;
#endif
	return(ticcount);
}

/*
============================================================================

								USER INPUT

============================================================================
*/

void MN_InitResolutionSpecificVars(void);
void R_InitResolutionSpecificVars(void);
void SB_InitResolutionSpecificVars(void);
void R_InitSky(int map);
void R_InitLightTables(void);
void R_ExecuteSetViewSize(void);

extern int screenblocks;
extern int detailLevel;
extern int resolutionX;
extern int resolutionY;

static void I_UpdateAllResolutionSpecificVariables(void)
{
	MN_InitResolutionSpecificVars();
	V_Init();
	BorderNeedRefresh = true;
	SB_state = -1;
	R_InitSky(gamemap);
	R_InitLightTables();
	R_InitResolutionSpecificVars();
	SB_InitResolutionSpecificVars();
	R_ExecuteSetViewSize();
}


void I_ChangeResolution(void)
{
	DDSURFACEDESC2 ddsd;
	HRESULT hr;

	setresneeded = false;
	g_ScreenWidth = resolutionX;
	g_ScreenHeight = resolutionY;

	I_UpdateAllResolutionSpecificVariables();

	if(g_pDDSScreen)
	{
		IDirectDrawSurface4_SetPalette(g_pDDS, NULL);
		IDirectDrawSurface4_Release(g_pDDSScreen);
	}

	if(g_pDDS)
	{
		IDirectDrawSurface4_SetPalette(g_pDDS, NULL);
		IDirectDrawSurface4_Release(g_pDDS);
	}

	if(FAILED(hr = IDirectDraw4_SetDisplayMode(g_pDD, g_ScreenWidth, g_ScreenHeight, 8, 0, 
											  (g_ScreenWidth == 320 && g_ScreenHeight == 200 ? DDSDM_STANDARDVGAMODE : 0))))
	{
		I_Error("Could not set a fullscreen resolution of %dx%d (Error code 0x%08x)", g_ScreenWidth, g_ScreenHeight, hr);
	}

	ZeroMemory(&ddsd, sizeof(DDSURFACEDESC2));

	ddsd.dwSize = sizeof(DDSURFACEDESC2);
	ddsd.dwFlags = DDSD_CAPS;
	ddsd.ddsCaps.dwCaps = DDSCAPS_PRIMARYSURFACE;

	if(FAILED(hr = IDirectDraw4_CreateSurface(g_pDD, &ddsd, &g_pDDS, NULL)))
	{
		I_Error("Could not create DirectDraw Surface! (Error code 0x%08x)", hr);
	}

	if(FAILED(hr = IDirectDrawSurface4_SetPalette(g_pDDS, g_pDDP)))
	{
		I_Error("Could not set Palette! (Error code 0x%08x)", hr);
	}

	ZeroMemory(&ddsd, sizeof(DDSURFACEDESC2));
	ZeroMemory(&ddsd.ddpfPixelFormat, sizeof(DDPIXELFORMAT));
	ddsd.dwSize = sizeof(DDSURFACEDESC2);
	ddsd.dwFlags = DDSD_WIDTH | DDSD_HEIGHT | DDSD_LPSURFACE |
                   DDSD_PITCH | DDSD_PIXELFORMAT | DDSD_CAPS;
	ddsd.dwWidth = g_ScreenWidth;
	ddsd.dwHeight = g_ScreenHeight;
	ddsd.lPitch = g_ScreenWidth;
	ddsd.lpSurface = screen;
	ddsd.ddsCaps.dwCaps = DDSCAPS_OFFSCREENPLAIN | DDSCAPS_SYSTEMMEMORY;
	ddsd.ddpfPixelFormat.dwSize = sizeof(DDPIXELFORMAT);
	ddsd.ddpfPixelFormat.dwFlags = DDPF_PALETTEINDEXED8 | DDPF_RGB;
	ddsd.ddpfPixelFormat.dwRGBBitCount = 8;

	if(FAILED(hr = IDirectDraw4_CreateSurface(g_pDD, &ddsd, &g_pDDSScreen, NULL)))
	{
		I_Error("Could not create DirectDrawSurface on 'screen' (Error code 0x%08x)", hr);
	}
}


//--------------------------------------------------------------------------
//
// PROC I_WaitVBL
//
//--------------------------------------------------------------------------

void I_WaitVBL(int vbls)
{
	int stat;

	if(novideo)
	{
		return;
	}

	_ASSERTE(g_pDD);

	for(stat = 0; stat < vbls; stat++)
	{
		IDirectDraw4_WaitForVerticalBlank(g_pDD, DDWAITVB_BLOCKEND, NULL);
		IDirectDraw4_WaitForVerticalBlank(g_pDD, DDWAITVB_BLOCKBEGIN, NULL);
	}
}

//--------------------------------------------------------------------------
//
// PROC I_SetPalette
//
// Palette source must use 8 bit RGB elements.
//
//--------------------------------------------------------------------------

void I_SetPalette(byte *palette)
{
	int i;
	BYTE *pal2 = palette;
	HRESULT hr;

	if(novideo)
	{
		return;
	}

	for(i = 0; i < 256; i++, pal2 += 3)
	{
		g_palLastSet[i].peRed = gammatable[usegamma][pal2[0]];
		g_palLastSet[i].peGreen = gammatable[usegamma][pal2[1]];
		g_palLastSet[i].peBlue = gammatable[usegamma][pal2[2]];
//		g_palLastSet[i].peRed = pal2[0];
//		g_palLastSet[i].peGreen = pal2[1];
//		g_palLastSet[i].peBlue = pal2[2];
	}

	_ASSERTE(g_pDDP);

//	I_WaitVBL(1);

	if(FAILED(hr = IDirectDrawPalette_SetEntries(g_pDDP, 0, 0, 256, g_palLastSet)))
	{
		I_Error("Could not set new palette! (Error code 0x%08x)", hr);
	}

//	if(FAILED(hr = IDirectDrawSurface4_SetPalette(g_pDDS, g_pDDP)))
//	{
//		I_Error("Could not set new palette! (Error code 0x%08x)", hr);
//	}

/*
	_outbyte(PEL_WRITE_ADR, 0);
	for(i = 0; i < 768; i++)
	{
		_outbyte(PEL_DATA, (gammatable[usegamma][*palette++])>>2);
	}
*/
}

void I_GetPalette(byte *rgPalBuf)
{
	int i;

	for(i = 0; i < 256; i++)
	{
		*rgPalBuf++ = g_palLastSet[i].peRed;
		*rgPalBuf++ = g_palLastSet[i].peGreen;
		*rgPalBuf++ = g_palLastSet[i].peBlue;
	}
}

/*
============================================================================

							GRAPHICS MODE

============================================================================
*/

byte *pcscreen, *destscreen, *destview;

HRESULT I_SmartDDSBlt(IDirectDrawSurface4 *pDDSDest, RECT *pRectDest, IDirectDrawSurface4 *pDDSSrc, RECT *pRectSrc, DWORD dwFlags, DDBLTFX *pBltFx)
{
	HRESULT hr;

	hr = IDirectDrawSurface4_Blt(pDDSDest, pRectDest, pDDSSrc, pRectSrc, dwFlags, pBltFx);

	if(hr == DDERR_SURFACELOST)
	{
		// We expect that 'pDDSDest' is always the primary surface.
		// If it was lost, we must also restore the original video mode...
		//
		hr = S_OK;

		if(SUCCEEDED(hr = IDirectDraw4_SetCooperativeLevel(g_pDD, g_hWnd, DDSCL_EXCLUSIVE | DDSCL_FULLSCREEN | DDSCL_ALLOWREBOOT)))
		{
			if(SUCCEEDED(hr = IDirectDraw4_SetDisplayMode(g_pDD, g_ScreenWidth, g_ScreenHeight, 8, 0, 0)))
			{
				if(SUCCEEDED(hr))
				{
					if(SUCCEEDED(hr = IDirectDraw4_RestoreAllSurfaces(g_pDD)))
					{
						if(SUCCEEDED(hr = IDirectDrawSurface4_SetPalette(pDDSDest, g_pDDP)))
						{
							// Restore the palette while I'm at it...
							//
							if(SUCCEEDED(hr = IDirectDrawPalette_SetEntries(g_pDDP, 0, 0, 256, g_palLastSet)))
							{
								// Finally, I am ready to try the Blt again...
								//
								hr = IDirectDrawSurface4_Blt(pDDSDest, pRectDest, pDDSSrc, pRectSrc, dwFlags, pBltFx);
							}
						}
					}
				}
			}
		}
	}
	else if(FAILED(hr))
	{
		ST_Message("Blt failed with error other than SURFACELOST (Error code 0x%08x)\n", hr);
	}

	return hr;
}


/*
==============
=
= I_Update
=
==============
*/

int UpdateState;
extern int screenblocks;

void I_Update (void)
{
	static int lasttic;
	HRESULT hr;
	RECT rectFull;

	rectFull.left = rectFull.top = 0;
	rectFull.right = g_ScreenWidth;
	rectFull.bottom = g_ScreenHeight;

	if(!g_pDD)
	{
		// In case DirectDraw was already shut down, don't update
		// the screen anymore...
		//
		return;
	}


//
// blit screen to video
//
	if(DisplayTicker)
	{
		byte *dest = screen;
		int i, tics;

/*
		if(screenblocks > 9 || UpdateState&(I_FULLSCRN|I_MESSAGES))
		{
			dest = (byte *)screen;
		}
		else
		{
			dest = (byte *)pcscreen;
		}
*/
		tics = ticcount-lasttic;
		lasttic = ticcount;
		if(tics > 20)
		{
			tics = 20;
		}
		for(i = 0; i < tics; i++)
		{
			*dest = 0xff;
			dest += 2;
		}
		for(i = tics; i < 20; i++)
		{
			*dest = 0x00;
			dest += 2;
		}
	}

	//memset(pcscreen, 255, SCREENHEIGHT*SCREENWIDTH);

	if(UpdateState == I_NOUPDATE)
	{
		return;
	}

//	ZeroMemory(&ddsd, sizeof(DDSURFACEDESC2));
//	ddsd.dwSize = sizeof(DDSURFACEDESC2);

/*
	if(FAILED(hr = IDirectDrawSurface4_Lock(g_pDDS, &rectToLock, &ddsd, DDLOCK_SURFACEMEMORYPTR, NULL)))
	{
		I_Error("Could not lock DirectDraw Surface! (Error code 0x%08x)", hr);
	}
*/
/*
	if(FAILED(hr = IDirectDrawSurface4_Lock(g_pDDS, NULL, &ddsd, DDLOCK_SURFACEMEMORYPTR, NULL)))
	{
		I_Error("Could not lock DirectDraw Surface! (Error code 0x%08x)", hr);
	}
*/
	// Point 'dest' at the visible screen...
	//
//	dest = ddsd.lpSurface;

	if(UpdateState&I_FULLSCRN)
	{
		if(FAILED(hr = I_SmartDDSBlt(g_pDDS, &rectFull, g_pDDSScreen, NULL, DDBLT_WAIT, NULL)))
		{
			I_Error("Could not blit buffer to visible screen! (Error code 0x%08x)", hr);
		}
/*
		for(i = 0; i < SCREENHEIGHT; i++, dest += ddsd.lPitch)
		{
			memcpy(dest, screen + (i * SCREENWIDTH), SCREENWIDTH);
		}
*/
//		memcpy(pcscreen, screen, SCREENWIDTH*SCREENHEIGHT);
		UpdateState = I_NOUPDATE; // clear out all draw types
	}
	if(UpdateState&I_FULLVIEW)
	{
		if(UpdateState&I_MESSAGES && screenblocks > 7)
		{
			RECT rect = { 0, 0, g_ScreenWidth, viewwindowy+viewheight };

			if(FAILED(hr = I_SmartDDSBlt(g_pDDS, &rect, g_pDDSScreen, &rect, DDBLT_WAIT, NULL)))
			{
				I_Error("Could not blit buffer to visible screen! (Error code 0x%08x)", hr);
			}
/*
			for(i = 0; i <
				(viewwindowy+viewheight)*SCREENWIDTH; i += SCREENWIDTH, dest += ddsd.lPitch)
			{
//				memcpy(pcscreen+i, screen+i, SCREENWIDTH);
				memcpy(dest, screen+i, SCREENWIDTH);
			}
*/
			UpdateState &= ~(I_FULLVIEW|I_MESSAGES);
		}
		else
		{
			RECT rect = { viewwindowx, viewwindowy, viewwindowx + viewwidth, viewwindowy + viewheight };

			if(FAILED(hr = I_SmartDDSBlt(g_pDDS, &rect, g_pDDSScreen, &rect, DDBLT_WAIT, NULL)))
			{
				I_Error("Could not blit buffer to visible screen! (Error code 0x%08x)", hr);
			}
/*
			for(i = viewwindowy; i < viewwindowy + viewheight; i++)
			{
				dest = ((BYTE *)ddsd.lpSurface) + (i * ddsd.lPitch) + viewwindowx;
				src = screen + (i * SCREENWIDTH) + viewwindowx;

				memcpy(dest, src, viewwidth);
			}
*/
/*
			for(i = viewwindowy*SCREENWIDTH+viewwindowx; i <
				(viewwindowy+viewheight)*SCREENWIDTH; i += SCREENWIDTH)
			{
				memcpy(pcscreen+i, screen+i, viewwidth);
			}
*/
			UpdateState &= ~I_FULLVIEW;
		}
	}
	if(UpdateState&I_STATBAR)
	{
		RECT rect = { 0, g_ScreenHeight - SBARHEIGHT, g_ScreenWidth, g_ScreenHeight };

		if(FAILED(hr = I_SmartDDSBlt(g_pDDS, &rect, g_pDDSScreen, &rect, DDBLT_WAIT, NULL)))
		{
			I_Error("Could not blit buffer to visible screen! (Error code 0x%08x)", hr);
		}
/*
		dest = ((BYTE *)ddsd.lpSurface) + (ddsd.lPitch * (SCREENHEIGHT - SBARHEIGHT));
		src = screen + (SCREENWIDTH * (SCREENHEIGHT - SBARHEIGHT));

		for(i = 0; i < SBARHEIGHT; i++, dest += ddsd.lPitch, src += SCREENWIDTH)
		{
			memcpy(dest, src, SCREENWIDTH);
		}
*/
/*
		memcpy(pcscreen+SCREENWIDTH*(SCREENHEIGHT-SBARHEIGHT),
			screen+SCREENWIDTH*(SCREENHEIGHT-SBARHEIGHT),
			SCREENWIDTH*SBARHEIGHT);
*/
		UpdateState &= ~I_STATBAR;
	}
	if(UpdateState&I_MESSAGES)
	{
		RECT rect = { 0, 0, g_ScreenWidth, 28 };

		if(FAILED(hr = I_SmartDDSBlt(g_pDDS, &rect, g_pDDSScreen, &rect, DDBLT_WAIT, NULL)))
		{
			I_Error("Could not blit buffer to visible screen! (Error code 0x%08x)", hr);
		}
/*
		dest = ddsd.lpSurface;
		src = screen;

		for(i = 0; i < 28; i++, dest += ddsd.lPitch, src += SCREENWIDTH)
		{
			memcpy(dest, src, SCREENWIDTH);
		}
//		memcpy(pcscreen, screen, SCREENWIDTH*28);
*/
		UpdateState &= ~I_MESSAGES;
	}

//	IDirectDrawSurface4_Unlock(g_pDDS, &rectToLock);
//	IDirectDrawSurface4_Unlock(g_pDDS, NULL);

//  memcpy(pcscreen, screen, SCREENHEIGHT*SCREENWIDTH);
}


// Display the current contents of 'screen' in the lower-left
// corner of the monitor...
//
void I_ShowCurrentContentsOfScreen(void)
{
	DDSURFACEDESC2 ddsd;

	ZeroMemory(&ddsd, sizeof(DDSURFACEDESC2));
	ddsd.dwSize = sizeof(DDSURFACEDESC2);

	if(g_pDDS && SUCCEEDED(IDirectDrawSurface4_Lock(g_pDDS, NULL, &ddsd, DDLOCK_SURFACEMEMORYPTR, NULL)))
	{
		DWORD y;
		BYTE *pbDest = (BYTE *)ddsd.lpSurface + ((ddsd.dwHeight - g_ScreenHeight) * ddsd.lPitch);

		for(y = 0; y < g_ScreenHeight; y++, pbDest += ddsd.lPitch)
		{
			memcpy(pbDest, screen + (y * g_ScreenWidth), g_ScreenWidth);
		}

		IDirectDrawSurface4_Unlock(g_pDDS, NULL);
	}
}


void I_ShowTextureContents(byte *pbBuffer, int width, int height)
{
	DDSURFACEDESC2 ddsd;

	ZeroMemory(&ddsd, sizeof(DDSURFACEDESC2));
	ddsd.dwSize = sizeof(DDSURFACEDESC2);

	if(g_pDDS && SUCCEEDED(IDirectDrawSurface4_Lock(g_pDDS, NULL, &ddsd, DDLOCK_SURFACEMEMORYPTR, NULL)))
	{
		DWORD y;
		BYTE *pbDest = (BYTE *)ddsd.lpSurface + ((ddsd.dwHeight - height) * ddsd.lPitch);

		for(y = 0; y < height; y++, pbDest += ddsd.lPitch)
		{
			memcpy(pbDest, pbBuffer + (y * width), width);
		}

		IDirectDrawSurface4_Unlock(g_pDDS, NULL);
	}
}


long I_StretchBlitToEntireScreen(byte *pbSrc)
{
	RECT rectDest;

	rectDest.left = rectDest.top = 0;
	rectDest.right = g_ScreenWidth;
	rectDest.bottom = g_ScreenHeight;

	if(g_pDD && g_pDDSScreen)
	{
		DDSURFACEDESC2 ddsd;
		IDirectDrawSurface4 *pDDS;
		HRESULT hr;

		ZeroMemory(&ddsd, sizeof(DDSURFACEDESC2));
		ddsd.dwSize = sizeof(DDSURFACEDESC2);
		ddsd.dwFlags = DDSD_WIDTH | DDSD_HEIGHT | DDSD_LPSURFACE |
					   DDSD_PITCH | DDSD_PIXELFORMAT | DDSD_CAPS;
		ddsd.dwWidth = 320;
		ddsd.dwHeight = 200;
		ddsd.lPitch = 320;
		ddsd.lpSurface = pbSrc;
		ddsd.ddsCaps.dwCaps = DDSCAPS_OFFSCREENPLAIN | DDSCAPS_SYSTEMMEMORY;
		ddsd.ddpfPixelFormat.dwSize = sizeof(DDPIXELFORMAT);
		ddsd.ddpfPixelFormat.dwFlags = DDPF_PALETTEINDEXED8 | DDPF_RGB;
		ddsd.ddpfPixelFormat.dwRGBBitCount = 8;

		if(FAILED(hr = IDirectDraw4_CreateSurface(g_pDD, &ddsd, &pDDS, NULL)))
		{
			return hr;
		}

		hr = IDirectDrawSurface4_Blt(g_pDDSScreen, &rectDest, pDDS, NULL, DDBLT_WAIT, NULL);

		IDirectDrawSurface4_Release(pDDS);

		return hr;
	}
	else
	{
		return E_FAIL;
	}
}

#if 0

static void DumpRaw8Bit(byte *pSrc, int width, int height, unsigned char *pal)
{
	int x, y;
	FILE *fp;
	unsigned char colorBuf[3];

	fp = fopen("raw8bit.raw", "wb");

	for(y = 0; y < height; y++)
	{
		byte *pSrcRow = pSrc + (y * width);

		for(x = 0; x < width; x++)
		{
			int SrcIndex = *(pSrcRow + x);
			byte *palOff = pal + (SrcIndex * 3);

			colorBuf[0] = *(palOff + 0);
			colorBuf[1] = *(palOff + 1);
			colorBuf[2] = *(palOff + 2);

			fwrite(colorBuf, 3, 1, fp);
		}
	}

	fclose(fp);
}


static void DumpRaw24Bit(byte *pSrc, int width, int height)
{
	FILE *fp;

	fp = fopen("raw24bit.raw", "wb");

	fwrite(pSrc, 1, width * height * 3, fp);

	fclose(fp);
}

static void DumpPalette(unsigned char *pal)
{
	FILE *fp;
	char szBuffer[80];
	int i;

	fp = fopen("hexpal.pal", "wt");

	fputs("JASC-PAL\n", fp);
	fputs("0100\n", fp);
	fputs("256\n", fp);

	for(i = 0; i < 256; i++)
	{
		sprintf(szBuffer, "%d %d %d\n", (long)*(pal + (i * 3)), (long)*(pal + (i * 3) + 1), (long)*(pal + (i * 3) + 2));
		fputs(szBuffer, fp);
	}

	fclose(fp);
}


static fixed_t ColorDist(fixed_t r1, fixed_t g1, fixed_t b1, fixed_t r2, fixed_t g2, fixed_t b2)
{
	int rDiff, gDiff, bDiff;
	int rDiff2, gDiff2, bDiff2;

	rDiff = (r1 - r2) >> FRACBITS;
	gDiff = (g1 - g2) >> FRACBITS;
	bDiff = (b1 - b2) >> FRACBITS;

	rDiff2 = rDiff * rDiff;
	gDiff2 = gDiff * gDiff;
	bDiff2 = bDiff * bDiff;

	return (3 * rDiff2)	+ (4 * gDiff2) + (2 * bDiff2);
}


static byte FindNearestColorTo(fixed_t r, fixed_t g, fixed_t b, fixed_t *fracpal)
{
	int i, j;
	fixed_t closestMatchValue;
	int closestMatchIndex;

	closestMatchValue = 0x7FFFFFFF;

	for(i = 0; i < 256; i++)
	{
		int diff;

		j = i * 3;

/*
		diff = abs(r - *(fracpal + j))		+
			   abs(g - *(fracpal + j + 1)	+
			   abs(b - *(fracpal + j + 2));
*/
		diff = ColorDist(r, g, b, *(fracpal + j), *(fracpal + j + 1), *(fracpal + j + 2));
		if(diff < closestMatchValue)
		{
			closestMatchValue = diff;
			closestMatchIndex = i;
			if(diff == 0)
			{
				break;
			}
		}
	}

	return closestMatchIndex;
}

static int DitherMap[8][8] = {  { 0, 32,  8, 40,  2, 34, 10, 42},
								{48, 16, 56, 24, 50, 18, 58, 26},
								{12, 44,  4, 36, 14, 46,  6, 38},
								{60, 28, 52, 20, 62, 30, 54, 22},
								{ 3, 35, 11, 43,  1, 33,  9, 41},
								{51, 19, 59, 27, 49, 17, 57, 25},
								{15, 47,  7, 39, 13, 45,  5, 37},
								{63, 31, 55, 23, 61, 29, 53, 21} };

void I_InterpolateStretchBlt(byte *pbDest, int DestWidth, int DestHeight, byte *pbSrc, int SrcWidth, int SrcHeight)
{
	unsigned char pal[256][3];
	int xDest, yDest, yDestOffset;
	int xSrc, ySrc, ySrcOffset;
	fixed_t u, v;
	fixed_t uDelta, vDelta;
	fixed_t xScale, yScale;
	fixed_t omuDelta, omvDelta;
	fixed_t *pDestTemp;
	byte *rgfHaveColorMapped;
	byte *rgbNearestColor;
	int fracpal[256][3];
	int i;

	I_GetPalette(&pal[0][0]);

	for(i = 0; i < 256; i++)
	{
		fracpal[i][0] = (int)pal[i][0] << FRACBITS;
		fracpal[i][1] = (int)pal[i][1] << FRACBITS;
		fracpal[i][2] = (int)pal[i][2] << FRACBITS;
	}

	pDestTemp = (fixed_t *)malloc((DestWidth * DestHeight * 3) * sizeof(fixed_t));

	xScale = FixedDiv(SrcWidth << FRACBITS, DestWidth << FRACBITS);
	yScale = FixedDiv(SrcHeight << FRACBITS, DestHeight << FRACBITS);

	for(yDest = 0; yDest < DestHeight; yDest++)
	{
		v = FixedMul(yDest << FRACBITS, yScale);
		ySrc = v >> FRACBITS;
		vDelta = v & 0xFFFF;
		omvDelta = FRACUNIT - vDelta;
		yDestOffset = (yDest * DestWidth);
		ySrcOffset = (ySrc * SrcWidth);

		for(xDest = 0; xDest < DestWidth; xDest++)
		{
			int r1, r2, r3, r4;
			int g1, g2, g3, g4;
			int b1, b2, b3, b4;
			int rFin, gFin, bFin;
			int temp1, temp2, temp3, temp4;
			byte *pHomeSrc;
			fixed_t *pHomeDest;

			u = FixedMul(xDest << FRACBITS, xScale);
			xSrc = u >> FRACBITS;
			uDelta = u & 0xFFFF;
			omuDelta = FRACUNIT - uDelta;

			temp1 = FixedMul(omuDelta, omvDelta);
			temp2 = FixedMul(uDelta, omvDelta);
			temp3 = FixedMul(omuDelta, vDelta);
			temp4 = FixedMul(uDelta, vDelta);

			pHomeSrc = pbSrc + ySrcOffset + xSrc;

			// Don't interpolate from or two black (color 0)!
			//
			if(*pHomeSrc)
			{
				r1 = fracpal[*pHomeSrc][0];
				g1 = fracpal[*pHomeSrc][1];
				b1 = fracpal[*pHomeSrc][2];

				if(xSrc < SrcWidth - 1 && *(pHomeSrc + 1))
				{
					r2 = fracpal[*(pHomeSrc + 1)][0];
					g2 = fracpal[*(pHomeSrc + 1)][1];
					b2 = fracpal[*(pHomeSrc + 1)][2];
				}
				else
				{
					r2 = r1;
					g2 = g1;
					b2 = b1;
				}

				if(ySrc < SrcHeight - 1 && *(pHomeSrc + SrcWidth))
				{
					r3 = fracpal[*(pHomeSrc + SrcWidth)][0];
					g3 = fracpal[*(pHomeSrc + SrcWidth)][1];
					b3 = fracpal[*(pHomeSrc + SrcWidth)][2];
				}
				else
				{
					r3 = r1;
					g3 = g1;
					b3 = b1;
				}

				if(xSrc < SrcWidth - 1 && ySrc < SrcHeight - 1 && *(pHomeSrc + SrcWidth + 1))
				{
					r4 = fracpal[*(pHomeSrc + SrcWidth + 1)][0];
					g4 = fracpal[*(pHomeSrc + SrcWidth + 1)][1];
					b4 = fracpal[*(pHomeSrc + SrcWidth + 1)][2];
				}
				else
				{
					r4 = r1;
					g4 = g1;
					b4 = b1;
				}

				rFin = FixedMul(r1, temp1) + 
					   FixedMul(r2, temp2) + 
					   FixedMul(r3, temp3) + 
					   FixedMul(r4, temp4);
				gFin = FixedMul(g1, temp1) + 
					   FixedMul(g2, temp2) + 
					   FixedMul(g3, temp3) + 
					   FixedMul(g4, temp4);
				bFin = FixedMul(b1, temp1) + 
					   FixedMul(b2, temp2) + 
					   FixedMul(b3, temp3) + 
					   FixedMul(b4, temp4);
			}
			else
			{
				rFin = fracpal[*pHomeSrc][0];
				gFin = fracpal[*pHomeSrc][1];
				bFin = fracpal[*pHomeSrc][2];
			}

			pHomeDest = pDestTemp + ((yDestOffset + xDest) * 3);

			*(pHomeDest) = rFin;
			*(pHomeDest + 1) = gFin;
			*(pHomeDest + 2) = bFin;
		}
	}

//	DumpRaw24Bit(pDestTemp, DestWidth, DestHeight);

	rgfHaveColorMapped = (byte *)malloc(262144 / 8);
	rgbNearestColor = (byte *)malloc(262144);

	ZeroMemory(rgfHaveColorMapped, 262144 / 8);

	// Make an integer copy of the palette table, for
	// faster operation later...
	//
//	for(i = 0; i < 256; i++)
//	{
//		fracpal[i][0] = pal[i][0];
//		fracpal[i][1] = pal[i][1];
//		fracpal[i][2] = pal[i][2];
//	}

	// Ok, pDestTemp has our true-color bitmap. Now we need to map the colors...
	//
	for(yDest = 0; yDest < DestHeight; yDest++)
	{
		int yDestOffset = yDest * DestWidth;
		int yDestBelow = yDestOffset + DestWidth;

		for(xDest = 0; xDest < DestWidth; xDest++)
		{
			int trueDestOffset = (yDestOffset + xDest) * 3;
			int fRed, fGreen, fBlue;
			int colMapIndex;
			byte fColBit;
			int ColByte;
			int errRed, errGreen, errBlue;
			int oppRed, oppGreen, oppBlue;
			int oppMapIndex;

			fRed = *(pDestTemp + trueDestOffset);
			fGreen = *(pDestTemp + trueDestOffset + 1);
			fBlue = *(pDestTemp + trueDestOffset + 2);

			colMapIndex = ((fRed >> (FRACBITS + 2)) << 12)	|
						  ((fGreen >> (FRACBITS + 2)) << 6) |
						  (fBlue >> (FRACBITS + 2));

			ColByte = colMapIndex >> 3;
			fColBit = 1 << (colMapIndex & 0x07);

			if(!(rgfHaveColorMapped[ColByte] & fColBit))
			{
				rgbNearestColor[colMapIndex] = FindNearestColorTo(fRed, fGreen, fBlue, &fracpal[0][0]);
				rgfHaveColorMapped[ColByte] |= fColBit;
			}

			oppRed = min(FRACUNIT * 255, max(0, (fracpal[rgbNearestColor[colMapIndex]][0] << 1) - fRed));
			oppGreen = min(FRACUNIT * 255, max(0, (fracpal[rgbNearestColor[colMapIndex]][1] << 1) - fGreen));
			oppBlue = min(FRACUNIT * 255, max(0, (fracpal[rgbNearestColor[colMapIndex]][2] << 1) - fBlue));
/*
			oppRed = min(FRACUNIT * 255, max(0, (fRed << 1) - fracpal[rgbNearestColor[colMapIndex]][0]));
			oppGreen = min(FRACUNIT * 255, max(0, (fGreen << 1) - fracpal[rgbNearestColor[colMapIndex]][1]));
			oppBlue = min(FRACUNIT * 255, max(0, (fBlue << 1) - fracpal[rgbNearestColor[colMapIndex]][2]));
*/

			oppMapIndex = ((oppRed >> (FRACBITS + 2)) << 12)	|
						  ((oppGreen >> (FRACBITS + 2)) << 6)	|
						  (oppBlue >> (FRACBITS + 2));
			ColByte = oppMapIndex >> 3;
			fColBit = 1 << (oppMapIndex & 0x07);

			if(!(rgfHaveColorMapped[ColByte] & fColBit))
			{
				rgbNearestColor[oppMapIndex] = FindNearestColorTo(oppRed, oppGreen, oppBlue, &fracpal[0][0]);
				rgfHaveColorMapped[ColByte] |= fColBit;
			}

			if(rgbNearestColor[oppMapIndex] != rgbNearestColor[colMapIndex])
			{
				fixed_t ditherConst;
				fixed_t trueToNearestDist;
				fixed_t oppToNearestDist;

				trueToNearestDist = ColorDist(fRed, fGreen, fBlue, 
					                          fracpal[rgbNearestColor[colMapIndex]][0],
											  fracpal[rgbNearestColor[colMapIndex]][1],
											  fracpal[rgbNearestColor[colMapIndex]][2]);

				oppToNearestDist = ColorDist(fracpal[rgbNearestColor[colMapIndex]][0],
											 fracpal[rgbNearestColor[colMapIndex]][1],
											 fracpal[rgbNearestColor[colMapIndex]][2],
					                         fracpal[rgbNearestColor[oppMapIndex]][0],
											 fracpal[rgbNearestColor[oppMapIndex]][1],
											 fracpal[rgbNearestColor[oppMapIndex]][2]);

				ditherConst = min(63, (64 * trueToNearestDist) / oppToNearestDist);
//				ditherConst >>= FRACBITS;

				assert(ditherConst >= 0 && ditherConst < 64);

				if(DitherMap[yDest & 0x07][xDest & 0x07] < ditherConst)
				{
					*(pbDest + yDestOffset + xDest) = rgbNearestColor[oppMapIndex];
				}
				else
				{
					*(pbDest + yDestOffset + xDest) = rgbNearestColor[colMapIndex];
				}
			}
			else
			{
				*(pbDest + yDestOffset + xDest) = rgbNearestColor[colMapIndex];
			}


#if 0	
			// Old FS-dither (broken?!)

			errRed = fracpal[rgbNearestColor[colMapIndex]][0] - fRed;
			errGreen = fracpal[rgbNearestColor[colMapIndex]][1] - fGreen;
			errBlue = fracpal[rgbNearestColor[colMapIndex]][2] - fBlue;

			if((errRed || errGreen || errBlue)	&& 
				yDest < DestHeight - 1			&& 
				xDest < DestWidth - 1)
			{
				// Distribute the error!
				//
				int rightDestOffset = trueDestOffset + 3;
				int belowDestOffset = (yDestBelow + xDest) * 3;
				int belowrightDestOffset = (yDestBelow + xDest + 1) * 3;
				fixed_t rightRed, belowRed, belowrightRed;
				fixed_t rightGreen, belowGreen, belowrightGreen;
				fixed_t rightBlue, belowBlue, belowrightBlue;
				fixed_t oneEighthRed, oneEighthGreen, oneEighthBlue;
				fixed_t threeEighthsRed, threeEighthsGreen, threeEighthsBlue;

				oneEighthRed = errRed >> 3;
				oneEighthGreen = errGreen >> 3;
				oneEighthBlue = errBlue >> 3;

				threeEighthsRed = oneEighthRed + oneEighthRed + oneEighthRed;
				threeEighthsGreen = oneEighthGreen + oneEighthGreen + oneEighthGreen;
				threeEighthsBlue = oneEighthBlue + oneEighthBlue + oneEighthBlue;

				rightRed = *(pDestTemp + rightDestOffset);
				rightGreen = *(pDestTemp + rightDestOffset + 1);
				rightBlue = *(pDestTemp + rightDestOffset + 2);

				belowRed = *(pDestTemp + belowDestOffset);
				belowGreen = *(pDestTemp + belowDestOffset + 1);
				belowBlue = *(pDestTemp + belowDestOffset + 2);

				belowrightRed = *(pDestTemp + belowrightDestOffset);
				belowrightGreen = *(pDestTemp + belowrightDestOffset + 1);
				belowrightBlue = *(pDestTemp + belowrightDestOffset + 2);

				rightRed += threeEighthsRed;
				rightGreen += threeEighthsGreen;
				rightBlue += threeEighthsBlue;

				belowRed += threeEighthsRed;
				belowGreen += threeEighthsGreen;
				belowBlue += threeEighthsBlue;

				belowrightRed += errRed >> 2;
				belowrightGreen += errGreen >> 2;
				belowrightBlue += errBlue >> 2;

#define CLAMP_BYTE(x)	{									\
							if((x) < 0)						\
								(x) = 0;					\
							else if((x) > FRACUNIT * 255)	\
								(x) = FRACUNIT * 255;		\
						}

				CLAMP_BYTE(rightRed);
				CLAMP_BYTE(belowRed);
				CLAMP_BYTE(belowrightRed);

				CLAMP_BYTE(rightGreen);
				CLAMP_BYTE(belowGreen);
				CLAMP_BYTE(belowrightGreen);

				CLAMP_BYTE(rightBlue);
				CLAMP_BYTE(belowBlue);
				CLAMP_BYTE(belowrightBlue);

				*(pDestTemp + rightDestOffset) = rightRed;
				*(pDestTemp + rightDestOffset + 1) = rightGreen;
				*(pDestTemp + rightDestOffset + 2) = rightBlue;

				*(pDestTemp + belowDestOffset) = belowRed;
				*(pDestTemp + belowDestOffset + 1) = belowGreen;
				*(pDestTemp + belowDestOffset + 2) = belowBlue;

				*(pDestTemp + belowrightDestOffset) = belowrightRed;
				*(pDestTemp + belowrightDestOffset + 1) = belowrightGreen;
				*(pDestTemp + belowrightDestOffset + 2) = belowrightBlue;
			}

			*(pbDest + yDestOffset + xDest) = rgbNearestColor[colMapIndex];

#endif	// 0

		}
	}

	free(rgfHaveColorMapped);
	free(rgbNearestColor);
	free(pDestTemp);
}

#endif	//0

void I_DDStretchBlt(byte *pbDest, int DestWidth, int DestHeight, byte *pbSrc, int SrcWidth, int SrcHeight)
{
	DDSURFACEDESC2 ddsd;
	IDirectDrawSurface4 *pDDSOrig;
	IDirectDrawSurface4 *pDDSScaled;
	HRESULT hr;

//	I_InterpolateStretchBlt(pbDest, DestWidth, DestHeight, pbSrc, SrcWidth, SrcHeight);
//	return;

	if(!g_pDD)
	{
		return;
	}

	ZeroMemory(&ddsd, sizeof(DDSURFACEDESC2));
	ddsd.dwSize = sizeof(DDSURFACEDESC2);
	ddsd.dwFlags = DDSD_WIDTH | DDSD_HEIGHT | DDSD_LPSURFACE |
				   DDSD_PITCH | DDSD_PIXELFORMAT | DDSD_CAPS;
	ddsd.dwWidth = SrcWidth;
	ddsd.dwHeight = SrcHeight;
	ddsd.lPitch = SrcWidth;
	ddsd.lpSurface = pbSrc;
	ddsd.ddsCaps.dwCaps = DDSCAPS_OFFSCREENPLAIN | DDSCAPS_SYSTEMMEMORY;
	ddsd.ddpfPixelFormat.dwSize = sizeof(DDPIXELFORMAT);
	ddsd.ddpfPixelFormat.dwFlags = DDPF_PALETTEINDEXED8 | DDPF_RGB;
	ddsd.ddpfPixelFormat.dwRGBBitCount = 8;

	if(FAILED(hr = IDirectDraw4_CreateSurface(g_pDD, &ddsd, &pDDSOrig, NULL)))
	{
		I_Error("Could not create DirectDrawSurface for Skies! (Error code 0x%08x)", hr);
	}

	ddsd.dwWidth = DestWidth;
	ddsd.dwHeight = DestHeight;
	ddsd.lPitch = DestWidth;
	ddsd.lpSurface = pbDest;

	if(FAILED(hr = IDirectDraw4_CreateSurface(g_pDD, &ddsd, &pDDSScaled, NULL)))
	{
		I_Error("Could not create DirectDrawSurface for Skies! (Error code 0x%08x)", hr);
	}

	if(FAILED(hr = IDirectDrawSurface4_Blt(pDDSScaled, NULL, pDDSOrig, NULL, DDBLT_WAIT, NULL)))
	{
		I_Error("Could not blt to Sky surface! (Error code 0x%08x)", hr);
	}

	IDirectDrawSurface4_Release(pDDSScaled);
	IDirectDrawSurface4_Release(pDDSOrig);
}


//--------------------------------------------------------------------------
//
// PROC I_InitGraphics
//
//--------------------------------------------------------------------------

void I_InitGraphics(void)
{
	HRESULT hr = S_OK;
	IDirectDraw *pDD = NULL;
	DDSURFACEDESC2 ddsd;
	HDC hdc;
	int i;

	if(novideo)
	{
		return;
	}

	if(!g_pDD)
	{
		// Ok, now setup DirectDraw...
		//
		if(FAILED(hr = DirectDrawCreate(NULL, &pDD, NULL)))
		{
			I_Error("Could not initialize DirectDraw! (Error code 0x%08x)", hr);
		}

		hr = IDirectDraw_QueryInterface(pDD, &IID_IDirectDraw4, (void **)&g_pDD);
		IDirectDraw_Release(pDD);

		if(FAILED(hr))
		{
			I_Error("Could not initialize DirectDraw! (Error code 0x%08x)", hr);
		}

		// Enumerate all availble video modes for the menu later...
		//
		ZeroMemory(&ddsd, sizeof(DDSURFACEDESC2));

		ddsd.dwSize = sizeof(DDSURFACEDESC2);
		ddsd.dwFlags = DDSD_PIXELFORMAT;
		ddsd.ddpfPixelFormat.dwSize = sizeof(DDPIXELFORMAT);
		ddsd.ddpfPixelFormat.dwFlags = DDPF_PALETTEINDEXED8 | DDPF_RGB;
		ddsd.ddpfPixelFormat.dwRGBBitCount = 8;

		if(FAILED(hr = IDirectDraw4_EnumDisplayModes(g_pDD, DDEDM_STANDARDVGAMODES, &ddsd, NULL, EnumModesCallback2)))
		{
			I_Error("Could not retrieve list of available video modes! (Error code 0x%08x)", hr);
		}

		if(FAILED(hr = IDirectDraw4_SetCooperativeLevel(g_pDD, g_hWnd, DDSCL_EXCLUSIVE | DDSCL_FULLSCREEN | DDSCL_ALLOWREBOOT)))
		{
			I_Error("Could not SetCooperativeLevel! (Error code 0x%08x)", hr);
		}
		if(FAILED(hr = IDirectDraw4_SetDisplayMode(g_pDD, g_ScreenWidth, g_ScreenHeight, 8, 0, 
												  (g_ScreenWidth == 320 && g_ScreenHeight == 200 ? DDSDM_STANDARDVGAMODE : 0))))
		{
			I_Error("Could not set a fullscreen resolution of %dx%d (Error code 0x%08x)", g_ScreenWidth, g_ScreenHeight, hr);
		}
	}
	else
	{
		// We already have the DirectDraw object created in the HR (High-Res) routines.
		//
		// We just need to set a new Display mode.
		//
		if(g_ScreenWidth != 640 || g_ScreenHeight != 480)
		{
			if(FAILED(hr = IDirectDraw4_SetDisplayMode(g_pDD, g_ScreenWidth, g_ScreenHeight, 8, 0, 
													  (g_ScreenWidth == 320 && g_ScreenHeight == 200 ? DDSDM_STANDARDVGAMODE : 0))))
			{
				I_Error("Could not set a fullscreen resolution of %dx%d (Error code 0x%08x)", g_ScreenWidth, g_ScreenHeight, hr);
			}
		}
	}

	ZeroMemory(&ddsd, sizeof(DDSURFACEDESC2));

	ddsd.dwSize = sizeof(DDSURFACEDESC2);
	ddsd.dwFlags = DDSD_CAPS;
//	ddsd.dwWidth = SCREENWIDTH;
//	ddsd.dwHeight = SCREENHEIGHT;
	ddsd.ddsCaps.dwCaps = DDSCAPS_PRIMARYSURFACE;

	if(FAILED(hr = IDirectDraw4_CreateSurface(g_pDD, &ddsd, &g_pDDS, NULL)))
	{
		I_Error("Could not create DirectDraw Surface! (Error code 0x%08x)", hr);
	}

	if(!g_pDDP)
	{
		hdc = GetDC(NULL);

		GetSystemPaletteEntries(hdc, 0, 256, g_palOriginal);
		SetSystemPaletteUse(hdc, SYSPAL_NOSTATIC);

		ReleaseDC(NULL, hdc);

		memcpy(g_palLastSet, g_palOriginal, 256 * sizeof(PALETTEENTRY));

		for (i = 0; i < 256; i++) 
			g_palLastSet[i].peFlags = PC_NOCOLLAPSE | PC_RESERVED;

//		g_palLastSet[0].peFlags = PC_EXPLICIT;
//		g_palLastSet[0].peRed = 0;
//		g_palLastSet[255].peFlags = PC_EXPLICIT;
//		g_palLastSet[255].peRed = 255;

		if(FAILED(hr = IDirectDraw4_CreatePalette(g_pDD, 
												  DDPCAPS_8BIT | DDPCAPS_ALLOW256,
												  g_palLastSet, 
												  &g_pDDP, 
												  NULL)))
		{
			I_Error("Could not create DirectDraw Palette! (Error code 0x%08x)", hr);
		}
	}

	if(FAILED(hr = IDirectDrawSurface4_SetPalette(g_pDDS, g_pDDP)))
	{
		I_Error("Could not set Palette! (Error code 0x%08x)", hr);
	}

	//grmode = true;
//	regs.w.ax = 0x13;
//	int386(0x10, (const union REGS *)&regs, &regs);

	// Acess-violate if these are used without LOCKing the DirectDraw surface...
	//
//	pcscreen = destscreen = (byte *)0xa0000;
	pcscreen = destscreen = destview = (byte *)0x00;

	SB_PaletteFlash(true);

/*
	I_SetPalette(W_CacheLumpName("PLAYPAL", PU_CACHE));

	// Wait a full refresh before setting the palette again...
	//
	hr = IDirectDraw4_WaitForVerticalBlank(g_pDD, DDWAITVB_BLOCKBEGIN, NULL);
	hr = IDirectDraw4_WaitForVerticalBlank(g_pDD, DDWAITVB_BLOCKEND, NULL);
	hr = IDirectDraw4_WaitForVerticalBlank(g_pDD, DDWAITVB_BLOCKBEGIN, NULL);

	if(FAILED(hr = IDirectDrawPalette_SetEntries(g_pDDP, 0, 0, 256, g_palLastSet)))
	{
		I_Error("Could not set new palette! (Error code 0x%08x)", hr);
	}
*/
	// Finally, wrap up 'screen' in a DirectDrawSurface object...
	//
	ZeroMemory(&ddsd, sizeof(DDSURFACEDESC2));
	ZeroMemory(&ddsd.ddpfPixelFormat, sizeof(DDPIXELFORMAT));
	ddsd.dwSize = sizeof(DDSURFACEDESC2);
	ddsd.dwFlags = DDSD_WIDTH | DDSD_HEIGHT | DDSD_LPSURFACE |
                   DDSD_PITCH | DDSD_PIXELFORMAT | DDSD_CAPS;
	ddsd.dwWidth = g_ScreenWidth;
	ddsd.dwHeight = g_ScreenHeight;
	ddsd.lPitch = g_ScreenWidth;
	ddsd.lpSurface = screen;
	ddsd.ddsCaps.dwCaps = DDSCAPS_OFFSCREENPLAIN | DDSCAPS_SYSTEMMEMORY;
	ddsd.ddpfPixelFormat.dwSize = sizeof(DDPIXELFORMAT);
	ddsd.ddpfPixelFormat.dwFlags = DDPF_PALETTEINDEXED8 | DDPF_RGB;
	ddsd.ddpfPixelFormat.dwRGBBitCount = 8;

	if(FAILED(hr = IDirectDraw4_CreateSurface(g_pDD, &ddsd, &g_pDDSScreen, NULL)))
	{
		I_Error("Could not create DirectDrawSurface on 'screen' (Error code 0x%08x)", hr);
	}

	I_UpdateAllResolutionSpecificVariables();
}

//--------------------------------------------------------------------------
//
// PROC I_ShutdownGraphics
//
//--------------------------------------------------------------------------

void I_ShutdownGraphics(void)
{
	HDC hdc;

	// Shutdown the automapper, which now has some state...
	//
	AM_Shutdown();

	// Restore the old palette state...
	//
	if(g_pDDP && g_pDDS)
	{
		IDirectDrawPalette_SetEntries(g_pDDP, 0, 0, 256, g_palOriginal);
		IDirectDrawSurface4_SetPalette(g_pDDS, g_pDDP);
	}

	// Kill DirectDraw...
	//
	if(g_pDDS)
	{
		IDirectDrawSurface4_SetPalette(g_pDDS, NULL);
		IDirectDrawSurface4_SetClipper(g_pDDS, NULL);
		IDirectDrawSurface4_Release(g_pDDS);
		g_pDDS = NULL;
	}
	if(g_pDDP)
	{
		IDirectDrawPalette_Release(g_pDDP);
		g_pDDP = NULL;
	}
	if(g_pDDSScreen)
	{
		IDirectDrawSurface4_Release(g_pDDSScreen);
		g_pDDSScreen = NULL;
	}
/*
	if(g_pDDC)
	{
		IDirectDrawClipper_Release(g_pDDC);
	}
*/
	if(g_pDD)
	{
		IDirectDraw4_RestoreDisplayMode(g_pDD);
		IDirectDraw4_Release(g_pDD);
		g_pDD = NULL;
	}

	// Restore the old palette state...
	//
	hdc = GetDC(NULL);
	SetSystemPaletteUse(hdc, SYSPAL_STATIC);
	ReleaseDC(NULL, hdc);

	if(g_rgScreenWidth)
	{
		free(g_rgScreenWidth);
	}
	if(g_rgScreenHeight)
	{
		free(g_rgScreenHeight);
	}

/*
	byte mode;

	// don't reset mode if it didn't get set
	mode = *(byte *)0x449;
	if(mode == 0x12 || mode == 0x13)
	{
		regs.w.ax = 3;
		int386(0x10, &regs, &regs); // back to text mode
	}
*/
}

//--------------------------------------------------------------------------
//
// PROC I_ReadScreen
//
// Reads the screen currently displayed into a linear buffer.
//
//--------------------------------------------------------------------------

/*
void I_ReadScreen(byte *scr)
{
	memcpy(scr, screen, SCREENWIDTH*SCREENHEIGHT);
}
*/

//===========================================================================

/*
===================
=
= I_StartTic
=
// called by D_DoomLoop
// called before processing each tic in a frame
// can call D_PostEvent
// asyncronous interrupt functions should maintain private ques that are
// read by the syncronous functions to be converted into events
===================
*/


#define SC_UPARROW              0x48
#define SC_DOWNARROW    0x50
#define SC_LEFTARROW            0x4b
#define SC_RIGHTARROW   0x4d

void   I_StartTic (void)
{
//	int             k;
//	event_t ev;


	// Mouse info is "pushed" by the input thread directly with H2_PostEvent() now,
	// so polling is unnecessary.
	//
//	I_ReadMouse ();

//
// keyboard events
//
	// Keyboard info is "pushed" by the input thread directly with H2_PostEvent() now,
	// so the keyboard queue is always empty.
	//

/*
	while (kbdtail < kbdhead)
	{
		k = keyboardque[kbdtail&(KBDQUESIZE-1)];
		kbdtail++;

		// extended keyboard shift key bullshit
		if ( (k&0x7f)==SC_LSHIFT || (k&0x7f)==SC_RSHIFT )
		{
			if ( keyboardque[(kbdtail-2)&(KBDQUESIZE-1)]==0xe0 )
				continue;
			k &= 0x80;
			k |= SC_RSHIFT;
		}

		if (k==0xe0)
			continue;               // special / pause keys
		if (keyboardque[(kbdtail-2)&(KBDQUESIZE-1)] == 0xe1)
			continue;                               // pause key bullshit

		if (k==0xc5 && keyboardque[(kbdtail-2)&(KBDQUESIZE-1)] == 0x9d)
		{
			ev.type = ev_keydown;
			ev.data1 = KEY_PAUSE;
			H2_PostEvent (&ev);
			continue;
		}

		if (k&0x80)
			ev.type = ev_keyup;
		else
			ev.type = ev_keydown;
		k &= 0x7f;
		switch (k)
		{
		case SC_UPARROW:
			ev.data1 = KEY_UPARROW;
			break;
		case SC_DOWNARROW:
			ev.data1 = KEY_DOWNARROW;
			break;
		case SC_LEFTARROW:
			ev.data1 = KEY_LEFTARROW;
			break;
		case SC_RIGHTARROW:
			ev.data1 = KEY_RIGHTARROW;
			break;
		default:
			ev.data1 = scantokey[k];
			break;
		}
		H2_PostEvent (&ev);
	}
*/
}


/*
void   I_ReadKeys (void)
{
	int             k;


	while (1)
	{
	   while (kbdtail < kbdhead)
	   {
		   k = keyboardque[kbdtail&(KBDQUESIZE-1)];
		   kbdtail++;
		   printf ("0x%x\n",k);
		   if (k == 1)
			   I_Quit ();
	   }
	}
}
*/

/*
===============
=
= I_StartFrame
=
===============
*/

void I_StartFrame (void)
{
/*
	I_JoystickEvents ();
	if(useexterndriver)
	{
		DPMIInt(i_Vector);
	}
*/
}

/*
============================================================================

					TIMER INTERRUPT

============================================================================
*/

/*
void I_ColorBlack (int r, int g, int b)
{
_outbyte (PEL_WRITE_ADR,0);
_outbyte(PEL_DATA,r);
_outbyte(PEL_DATA,g);
_outbyte(PEL_DATA,b);
}
*/


/*
============================================================================

						KEYBOARD

============================================================================
*/

//void (__interrupt __far *oldkeyboardisr) () = NULL;

int lastpress;

/*
================
=
= I_KeyboardISR
=
================
*/

/*
void __interrupt I_KeyboardISR (void)
{
// Get the scan code

	keyboardque[kbdhead&(KBDQUESIZE-1)] = lastpress = _inbyte(0x60);
	kbdhead++;

// acknowledge the interrupt

	_outbyte(0x20,0x20);
}
*/


HRESULT I_SmartGetDeviceState(IDirectInputDevice *pDID, DWORD dwcbData, void *pvBuffer)
{
	HRESULT hr;

	hr = IDirectInputDevice_GetDeviceState(pDID, dwcbData, pvBuffer);

	if(hr == DIERR_INPUTLOST)
	{
		// We must have lost the Window focus?
		//
		if(g_fActive == false)
		{
			// Sit around and do nothing until we are active again...
			//

			SetThreadPriority(GetCurrentThread(), THREAD_PRIORITY_IDLE);

			while(g_fActive == false);

			SetThreadPriority(GetCurrentThread(), THREAD_PRIORITY_NORMAL);
		}

		// Re-acquire the device...
		//
		if(SUCCEEDED(hr = IDirectInputDevice_Acquire(pDID)))
		{
			hr = IDirectInputDevice_GetDeviceState(pDID, dwcbData, pvBuffer);
		}
	}

	return hr;
}


void TranslateDIKToHexen(DWORD dwState, DWORD dwKey)
{
	event_t ev;

	if(dwState & 0x80)
	{
		ev.type = ev_keydown;
	}
	else
	{
		ev.type = ev_keyup;
	}

	// We need to special-case only the following keys
	// (the table 'scantokey[]' translates the rest of
	// the keyboard values to their ASCII/Hexen
	// equivelents.)
	//

	switch(dwKey)
	{
		case DIK_UP:
			ev.data1 = KEY_UPARROW;
			break;

		case DIK_DOWN:
			ev.data1 = KEY_DOWNARROW;
			break;

		case DIK_LEFT:
			ev.data1 = KEY_LEFTARROW;
			break;

		case DIK_RIGHT:
			ev.data1 = KEY_RIGHTARROW;
			break;

		case DIK_RMENU:
			ev.data1 = KEY_RALT;
			break;

		case 0xC5:
//		case DIK_PAUSE:
			ev.data1 = KEY_PAUSE;
			break;

		case DIK_LSHIFT:
			ev.data1 = KEY_RSHIFT;
			break;

		case DIK_NUMPADENTER:
			ev.data1 = KEY_ENTER;
			break;

		case DIK_RCONTROL:
			ev.data1 = KEY_RCTRL;
			break;

		case DIK_HOME:
			ev.data1 = KEY_HOME;
			break;

		case DIK_PRIOR:
			ev.data1 = KEY_PGUP;
			break;

		case DIK_END:
			ev.data1 = KEY_END;
			break;

		case DIK_NEXT:
			ev.data1 = KEY_PGDN;
			break;

		case DIK_INSERT:
			ev.data1 = KEY_INS;
			break;

		case DIK_DELETE:
			ev.data1 = KEY_DEL;
			break;

		default:
			if(dwKey < 0x80)
			{
				ev.data1 = scantokey[dwKey];
			}
			else
			{
				// Just eat these funky keys...
				//
				return;
			}
			break;
	}

	H2_PostEvent(&ev);
}


unsigned __stdcall InputThreadProc(void *pvData)
{
	IDirectInput *		pDI = NULL;
	IDirectInputDevice *pDIDKbd = NULL;
	IDirectInputDevice *pDIDMse = NULL;
	HRESULT				hr = S_OK;
//	DIPROPDWORD			dipd;
	DIMOUSESTATE		dims;
	HANDLE				hReady = NULL;
	HANDLE				hTerminate = NULL;
	HANDLE				rghPossibleEvents[3] = { NULL, NULL, NULL };
	DWORD				i = 0;
	char				rgKbd[256];
	char				rgKbdLast[256];
//	DIDEVCAPS			didc;

	if(FAILED(hr = DirectInputCreate(g_hInst, DIRECTINPUT_VERSION, &pDI, NULL)))
	{
		goto threadError;
	}

	if(FAILED(hr = IDirectInput_CreateDevice(pDI, &GUID_SysKeyboard, &pDIDKbd, NULL)))
	{
		goto threadError;
	}

	if(FAILED(hr = IDirectInput_CreateDevice(pDI, &GUID_SysMouse, &pDIDMse, NULL)))
	{
		goto threadError;
	}

	if(FAILED(hr = IDirectInputDevice_SetDataFormat(pDIDKbd, &c_dfDIKeyboard)))
	{
		goto threadError;
	}

	if(FAILED(hr = IDirectInputDevice_SetDataFormat(pDIDMse, &c_dfDIMouse)))
	{
		goto threadError;
	}

/*
	ZeroMemory(&didc, sizeof(DIDEVCAPS));
	didc.dwSize = sizeof(DIDEVCAPS);
	if(FAILED(hr = IDirectInputDevice_GetCapabilities(pDIDKbd, &didc)))
	{
		goto threadError;
	}

	if(didc.dwFlags & DIDC_POLLEDDATAFORMAT)
	{
		// Fuck!
		//
		MessageBox(NULL, "Your keyboard requires polling!", "Fuck!", MB_OK);
	}
*/

//#ifdef _DEBUG
	if(FAILED(hr = IDirectInputDevice_SetCooperativeLevel(pDIDKbd, g_hWnd, DISCL_NONEXCLUSIVE | DISCL_FOREGROUND)))
	{
		goto threadError;
	}
	if(FAILED(hr = IDirectInputDevice_SetCooperativeLevel(pDIDMse, g_hWnd, DISCL_NONEXCLUSIVE | DISCL_FOREGROUND)))
	{
		goto threadError;
	}
//#else
//	if(FAILED(hr = IDirectInputDevice_SetCooperativeLevel(pDIDKbd, g_hWnd, DISCL_NONEXCLUSIVE | DISCL_FOREGROUND)))
//	{
//		goto threadError;
//	}
//	if(FAILED(hr = IDirectInputDevice_SetCooperativeLevel(pDIDMse, g_hWnd, DISCL_EXCLUSIVE | DISCL_FOREGROUND)))
//	{
//		goto threadError;
//	}
//#endif

	// Hexen uses relative-mouse coordinates, which is the default for DirectInput.
	// To change to absolute coordinates, uncomment--out this code...
	//
/*
	ZeroMemory(&dipd, sizeof(dipd));
	dipd.diph.dwSize = sizeof(DIPROPDWORD);
	dipd.diph.dwHeaderSize = sizeof(DIPROPHEADER);
	dipd.diph.dwHow = DIPH_DEVICE;
	dipd.dwData = DIPROPAXISMODE_ABS;

	if(FAILED(hr = IDirectInputDevice_SetProperty(pDIDMse, DIPROP_AXISMODE, &dipd.diph)))
	{
		goto threadError;
	}
*/

	if((hReady = OpenEvent(EVENT_ALL_ACCESS, FALSE, gc_szInputThreadReady)) == NULL)
	{
		goto threadError;
	}

	if((hTerminate = OpenEvent(EVENT_ALL_ACCESS, FALSE, gc_szInputThreadTerminate)) == NULL)
	{
		goto threadError;
	}

	if((rghPossibleEvents[0] = CreateEvent(NULL, FALSE, FALSE, gc_szDIKeyboardEvent)) == NULL)
	{
		goto threadError;
	}

	if((rghPossibleEvents[1] = CreateEvent(NULL, FALSE, FALSE, gc_szDIMouseEvent)) == NULL)
	{
		goto threadError;
	}

	// In case the boss-man tells me to die...
	//
	rghPossibleEvents[2] = hTerminate;

	if(FAILED(hr = IDirectInputDevice_SetEventNotification(pDIDKbd, rghPossibleEvents[0])))
	{
		goto threadError;
	}

	if(FAILED(hr = IDirectInputDevice_SetEventNotification(pDIDMse, rghPossibleEvents[1])))
	{
		goto threadError;
	}

	if(FAILED(hr = IDirectInputDevice_Acquire(pDIDKbd)))
	{
		goto threadError;
	}

	if(FAILED(hr = IDirectInputDevice_Acquire(pDIDMse)))
	{
		goto threadError;
	}

	if(FAILED(hr = IDirectInputDevice_GetDeviceState(pDIDKbd, 256, rgKbdLast)))
	{
		goto threadError;
	}

	// Ok, I'm ready to rock-n-roll!
	// Tell the boss-man I'm ready...
	//
	SetEvent(hReady);

	do
	{
		i = WaitForMultipleObjects(3, rghPossibleEvents, FALSE, INFINITE);

		if(i == WAIT_OBJECT_0)
		{
			DWORD j;
			BYTE *pbKbdLast;
			BYTE bKbdCur;

			// Aha! A keyboard event! Grab the latest & greatest...
			//
			if(FAILED(hr = I_SmartGetDeviceState(pDIDKbd, 256, rgKbd)))
			{
				OutputDebugString("Could not read Keyboard state!\n");
				goto threadError;
			}

			// Well this is just great. DirectInput told me the keyboard has been
			// updated, but I have no idea what's changed! Time to figure it out...
			//
			for(j = 0, pbKbdLast = rgKbdLast; j < 256; j++, pbKbdLast++)
			{
				bKbdCur = rgKbd[j];

				if(bKbdCur != *pbKbdLast)
				{
					TranslateDIKToHexen(bKbdCur, j);
					*pbKbdLast = bKbdCur;
				}
			}
		}
		else if(i == WAIT_OBJECT_0 + 1)
		{
			event_t ev;

			// Aha! A mouse event! Grab the latest & greatest...
			//
			if(FAILED(hr = I_SmartGetDeviceState(pDIDMse, sizeof(DIMOUSESTATE), &dims)))
			{
				goto threadError;
			}

			// Mouse is always present in Windows -- use this flag as an over-ride
			// instead (for example, if they pass "-nomouse" on the command-line)
			//
			if(mousepresent && usemouse)
			{
				ev.type = ev_mouse;
				ev.data1 = ((dims.rgbButtons[0] & 0x80) >> 7)	| 
						   ((dims.rgbButtons[1] & 0x80) >> 6)	|
						   ((dims.rgbButtons[2] & 0x80) >> 5);
				ev.data2 = (short)dims.lX;
				ev.data3 = -(short)dims.lY;
//				ev.data2 = (short)(dims.lX * 8);
//				ev.data3 = -(short)(dims.lY * 8);
				H2_PostEvent(&ev);
			}
		}

		while(g_fActive == false);

	} while(i != WAIT_OBJECT_0 + 2);

threadError:
	if(pDIDKbd)
	{
		IDirectInputDevice_Unacquire(pDIDKbd);
		IDirectInputDevice_SetEventNotification(pDIDKbd, NULL);
		IDirectInputDevice_Release(pDIDKbd);
	}
	if(pDIDMse)
	{
		IDirectInputDevice_Unacquire(pDIDMse);
		IDirectInputDevice_SetEventNotification(pDIDMse, NULL);
		IDirectInputDevice_Release(pDIDMse);
	}
	if(pDI)
	{
		IDirectInput_Release(pDI);
	}
	if(hReady)
	{
		CloseHandle(hReady);
	}
	if(hTerminate)
	{
		CloseHandle(hTerminate);
	}
	for(i = 0; i < 2; i++)
	{
		if(rghPossibleEvents[i])
		{
			CloseHandle(rghPossibleEvents[i]);
		}
	}

	_endthreadex(hr);
	return hr;
}


/*
===============
=
= I_StartupKeyboard
=
===============
*/

void I_StartupMouseAndKeyboard(void)
{
	HANDLE rghMultiple[2];
	DWORD dwRet;
	unsigned nThreadAddr;

	if((g_hInputThreadReady = CreateEvent(NULL, FALSE, FALSE, gc_szInputThreadReady)) == NULL)
	{
		I_Error("Could not create InputThreadReady event! (Error code 0x%08x)", GetLastError());
	}

	if((g_hInputThreadTerminate = CreateEvent(NULL, FALSE, FALSE, gc_szInputThreadTerminate)) == NULL)
	{
		I_Error("Could not create InputThreadTerminate event! (Error code 0x%08x)", GetLastError());
	}

	if((g_hInputThread = (HANDLE)_beginthreadex(NULL, 0, InputThreadProc, NULL, 0, &nThreadAddr)) == 0)
	{
		I_Error("Could not create Input thread! (Error code 0x%08x)", errno);
	}

	rghMultiple[0] = g_hInputThreadReady;
	rghMultiple[1] = g_hInputThread;

#ifdef _DEBUG
	dwRet = WaitForMultipleObjects(2, rghMultiple, FALSE, 1000000);
#else
	dwRet = WaitForMultipleObjects(2, rghMultiple, FALSE, 10000);
#endif

	// If it takes longer than 10 seconds for the input thread to get ready, 
	// or if the thread terminates itself, then something is very wrong...
	//
	if(dwRet != WAIT_OBJECT_0)
	{
		if(dwRet == WAIT_TIMEOUT)
		{
			// This is a dangerous call, but at this point we have no choice...
			//
			TerminateThread(g_hInputThread, -1);
		}
		CloseHandle(g_hInputThread);
		g_hInputThread = NULL;

		I_Error("DirectInput failed to initialize properly!");
	}

	mousepresent = 1;

	// Otherwise everything is ok and the InputThread is working as it should be...
	//
/*
#ifndef NOKBD
	oldkeyboardisr = _dos_getvect(KEYBOARDINT);
	_dos_setvect (0x8000 | KEYBOARDINT, I_KeyboardISR);
#endif
*/

//I_ReadKeys ();
}


void I_ShutdownMouseAndKeyboard (void)
{
	// Basically all we have to do here is tell the Input thread to kill itself.
	// It will cleanup everything it was using...
	//

	SetEvent(g_hInputThreadTerminate);

	// Now wait for him to die...
	//
	if(WaitForSingleObject(g_hInputThread, 5000) == WAIT_TIMEOUT)
	{
		// Trouble...the input thread didn't die!
		//
		TerminateThread(g_hInputThread, -1);
	}

	CloseHandle(g_hInputThread);
	g_hInputThread = NULL;
	CloseHandle(g_hInputThreadTerminate);
	g_hInputThreadTerminate = NULL;
	CloseHandle(g_hInputThreadReady);
	g_hInputThreadReady = NULL;
/*
	if (oldkeyboardisr)
		_dos_setvect (KEYBOARDINT, oldkeyboardisr);
	*(short *)0x41c = *(short *)0x41a;      // clear bios key buffer
*/
}


/*
============================================================================

							MOUSE

============================================================================
*/


#if 0

int I_ResetMouse (void)
{
	regs.w.ax = 0;                  // reset
	int386 (0x33, &regs, &regs);
	return regs.w.ax;
}


/*
================
=
= StartupMouse
=
================
*/


void I_StartupMouse (void)
{
   //
   // General mouse detection
   //
	mousepresent = 0;
	if ( M_CheckParm ("-nomouse") || !usemouse )
		return;

	if (I_ResetMouse () != 0xffff)
	{
		ST_Message("Mouse: not present\n");
		return;
	}
	ST_Message("Mouse: detected\n");
	mousepresent = 1;

}


/*
================
=
= ShutdownMouse
=
================
*/

void I_ShutdownMouse (void)
{
	// MSE_TODO:
	if (!mousepresent)
	  return;

	I_ResetMouse ();
}


/*
================
=
= I_ReadMouse
=
================
*/

void I_ReadMouse (void)
{
	event_t ev;

//
// mouse events
//
	if (!mousepresent)
		return;

	ev.type = ev_mouse;

	memset (&dpmiregs,0,sizeof(dpmiregs));
	dpmiregs.eax = 3;                               // read buttons / position
	DPMIInt (0x33);
	ev.data1 = dpmiregs.ebx;

	dpmiregs.eax = 11;                              // read counters
	DPMIInt (0x33);
	ev.data2 = (short)dpmiregs.ecx;
	ev.data3 = -(short)dpmiregs.edx;

	H2_PostEvent (&ev);
}

#endif	// 0

/*
============================================================================

					TIMER INTERRUPT

============================================================================
*/

//void (__interrupt __far *oldtimerisr) ();


/*
void IO_ColorBlack (int r, int g, int b)
{
_outbyte (PEL_WRITE_ADR,0);
_outbyte(PEL_DATA,r);
_outbyte(PEL_DATA,g);
_outbyte(PEL_DATA,b);
}
*/


/*
================
=
= IO_TimerISR
=
================
*/
/*
void CALLBACK IO_TimerProc(HWND hwnd, UINT msg, UINT idEvent, DWORD dwTime)
{
	ticcount++;
}
*/

/*
===============
=
= IO_StartupTimer
=
===============
*/

unsigned __stdcall TimerThreadProc(void *pvData)
{
	HANDLE hTermEvent;
	long lTicsPerSec = 1193180 / VBLCOUNTER;
	long lMillsecsPerTic = (long)((1.0 / (double)lTicsPerSec) * 1000.0);

	hTermEvent = OpenEvent(EVENT_ALL_ACCESS, FALSE, gc_szTTTEName);

	// This thread eats virtually no CPU...
	//
	while(WaitForSingleObject(hTermEvent, lMillsecsPerTic) == WAIT_TIMEOUT)
	{
		ticcount++;
	}

	CloseHandle(hTermEvent);

	_endthreadex(0);
	return 0;
}


void IO_StartupTimer (void)
{
	int nTimerThreadID;

	// We'll spawn a new thread to run the timer.
	// The reason for this is that Windows' timers need a message
	// pump to trigger the WM_TIMER events. But the Hexen code
	// will loop in place waiting for time to elapse, without
	// yielding control to our "main-thread" message pump and thus
	// the 'ticcount' variable never gets updated...
	//
	// We'll also use an Event (Timer-Thread-Termination-Event) to signal 
	// to the thread that is should terminate itself.
	//
	g_hTTTE = CreateEvent(NULL, FALSE, FALSE, gc_szTTTEName);

	if(!g_hTTTE)
	{
		I_Error("Could not create Event for timer thread! (Error code 0x%08x)", GetLastError());
	}

	g_hTimerThread = (HANDLE)_beginthreadex(NULL, 0, TimerThreadProc, NULL, 0, &nTimerThreadID);

	if(!g_hTimerThread)
	{
		I_Error("Could not spawn timer thread! (Error code 0x%08x)", g_hTimerThread);
	}

	// Because this thread performs timing functions, give it a slightly
	// higher priority...
	//
	SetThreadPriority(g_hTimerThread, THREAD_PRIORITY_ABOVE_NORMAL);

/*
	oldtimerisr = _dos_getvect(TIMERINT);

	_dos_setvect (0x8000 | TIMERINT, IO_TimerISR);
	IO_SetTimer0 (VBLCOUNTER);
*/
}


void IO_ShutdownTimer (void)
{
	// All we have to do is trigger our TTT-Event and the timer-thread
	// will self-terminate...
	//
	SetEvent(g_hTTTE);

	// Give up my time-slice...
	//
	Sleep(0);

	// Now we're done with the event, we can close it...
	//
	CloseHandle(g_hTTTE);
	g_hTTTE = NULL;
	CloseHandle(g_hTimerThread);
	g_hTimerThread = NULL;

/*
	if (oldtimerisr)
	{
		IO_SetTimer0 (0);              // back to 18.4 ips
		_dos_setvect (TIMERINT, oldtimerisr);
	}
*/
}


//===========================================================================


/*
===============
=
= I_Init
=
= hook interrupts and set graphics mode
=
===============
*/

void I_SetClientWindowSize(int width, int height)
{
	RECT rect;
	int xWinSize, yWinSize;

	do
	{
		SetWindowPos(g_hWnd, HWND_TOP, 0, 0, width, height, 0);

		GetClientRect(g_hWnd, &rect);

		xWinSize = width - rect.right;
		yWinSize = height - rect.bottom;

		width += xWinSize;
		height += xWinSize;
	} while(xWinSize || yWinSize);
}


void I_Init (void)
{
	extern void I_StartupTimer(void);

	novideo = M_CheckParm("novideo");

	// First, create a window.
	//
	if((g_hWnd = CreateWindowEx(WS_EX_TOPMOST, 
								"HexenWindowClass", 
								"Hexen", 
								WS_POPUP | WS_VISIBLE, 
								0, 0, 
								g_ScreenWidth, g_ScreenHeight,
								NULL, NULL,
								g_hInst, NULL)) == NULL)
	{
		I_Error("Could not create Window! (Error code %d)", GetLastError());
	}

	ST_Message("  S_Init... ");
	S_Init();

	ST_Message("  Starting up mouse and keyboard...\n");
	I_StartupMouseAndKeyboard();

	IO_StartupTimer();
	S_Start();
}


/*
===============
=
= I_Shutdown
=
= return to default system state
=
===============
*/

void I_Shutdown (void)
{
	I_ShutdownGraphics ();

	IO_ShutdownTimer ();

	S_ShutDown ();
	I_ShutdownMouseAndKeyboard();
}


/*
================
=
= I_Error
=
================
*/

void I_Error (char *error, ...)
{
//	union REGS regs;

	va_list argptr;
	char szBuffer[2048];

	D_QuitNetGame ();

	I_Shutdown ();

	DestroyWindow(g_hWnd);

//	regs.x.eax = 0x3;
//	int386(0x10, &regs, &regs);

	va_start (argptr,error);
	vsprintf (szBuffer, error,argptr);
	va_end (argptr);

	strcat(szBuffer, "\n");
	OutputDebugString(szBuffer);

	// Since this is an error case, we are probably going to be right
	// in the middle of the main code path. We've already destroyed all
	// of our globals, so any further main-path code executed is going
	// to crash horribly. Therefore, cop-out and just "exit"...
	//

	if(strcmp(error, ""))
	{
		MessageBox(NULL, szBuffer, "Error", MB_OK | MB_ICONEXCLAMATION);
	}

	exit(1);
}

//--------------------------------------------------------------------------
//
// I_Quit
//
// Shuts down net game, saves defaults, prints the exit text message,
// goes to text mode, and exits.
//
//--------------------------------------------------------------------------

void I_Quit(void)
{
	D_QuitNetGame();
	M_SaveDefaults();
	I_Shutdown();

	// Since this method is called outside the main code path
	// (i.e. it's called from within our WindowProc) we can
	// destroy all of our globals without concern. No main-path
	// code will be executed after this method.
	//
	DestroyWindow(g_hWnd);
	PostQuitMessage(0);

//	scr = (byte *)W_CacheLumpName("ENDTEXT", PU_CACHE);
/*
	memcpy((void *)0xb8000, scr, 80*25*2);
	regs.w.ax = 0x0200;
	regs.h.bh = 0;
	regs.h.dl = 0;
	regs.h.dh = 23;
	int386(0x10, (const union REGS *)&regs, &regs); // Set text pos
	_settextposition(24, 1);
*/
//	printf("\nHexen: Beyond Heretic\n");
}

/*
===============
=
= I_ZoneBase
=
===============
*/

byte *I_ZoneBase (int *size)
{
//	char meminfo[32];
	int heap;
	byte *ptr;
	MEMORYSTATUS memstat;

	GlobalMemoryStatus(&memstat);

/*
	memset (meminfo,0,sizeof(meminfo));
	segread(&segregs);
	segregs.es = segregs.ds;
	regs.w.ax = 0x500;      // get memory info
	regs.x.edi = (int)&meminfo;
	int386x( 0x31, &regs, &regs, &segregs );

	heap = meminfo[0];
	ST_Message ("  DPMI memory: 0x%x, ",heap);
*/
	ST_Message("Win32 memory: 0x%x Physical, 0x%x Virtual\n", memstat.dwAvailPhys, memstat.dwAvailVirtual);
	ST_Message ("Maxzone: 0x%x\n", maxzone);

	if(memstat.dwAvailPhys >= maxzone)
	{
		ptr = malloc(maxzone);
		heap = maxzone;
	}
	else
	{
		heap = memstat.dwAvailPhys + memstat.dwAvailVirtual;

		do
		{
			heap -= 0x10000;                // leave 64k alone
			if (heap > maxzone)
				heap = maxzone;
			ptr = malloc (heap);
		} while (!ptr);
	}

	ST_Message ("  0x%x allocated for zone, ", heap);
	ST_Message ("ZoneBase: 0x%X\n", (int)ptr);

	if (heap < 0x180000)
		I_Error ("  Insufficient DPMI memory!");

#if 0
	regs.w.ax = 0x501;      // allocate linear block
	regs.w.bx = heap>>16;
	regs.w.cx = heap&0xffff;
	int386( 0x31, &regs, &regs);
	if (regs.w.cflag)
		I_Error ("  Couldn't allocate DPMI memory!");

	block = (regs.w.si << 16) + regs.w.di;
#endif

	*size = heap;
	return ptr;
}

/*
=============
=
= I_AllocLow
=
=============
*/

byte *I_AllocLow (int length)
{
	byte    *mem;

	mem = malloc(length);

	if(!mem)
	{
		I_Error ("I_AllocLow: Alloc of %i failed", length);
	}
/*
	// DPMI call 100h allocates DOS memory
	segread(&segregs);
	regs.w.ax = 0x0100;          // DPMI allocate DOS memory
	regs.w.bx = (length+15) / 16;
	int386( DPMI_INT, &regs, &regs);
//      segment = regs.w.ax;
//   selector = regs.w.dx;
	if (regs.w.cflag != 0)
		I_Error ("I_AllocLow: DOS alloc of %i failed, %i free",
			length, regs.w.bx*16);


	mem = (void *) ((regs.x.eax & 0xFFFF) << 4);
*/

	memset (mem,0,length);
	return mem;
}

/*
============================================================================

						NETWORKING

============================================================================
*/

/* // FUCKED LINES
typedef struct
{
	char    priv[508];
} doomdata_t;
*/ // FUCKED LINES

#define DOOMCOM_ID              0x12345678l

/* // FUCKED LINES
typedef struct
{
	long    id;
	short   intnum;                 // DOOM executes an int to execute commands

// communication between DOOM and the driver
	short   command;                // CMD_SEND or CMD_GET
	short   remotenode;             // dest for send, set by get (-1 = no packet)
	short   datalength;             // bytes in doomdata to be sent

// info common to all nodes
	short   numnodes;               // console is allways node 0
	short   ticdup;                 // 1 = no duplication, 2-5 = dup for slow nets
	short   extratics;              // 1 = send a backup tic in every packet
	short   deathmatch;             // 1 = deathmatch
	short   savegame;               // -1 = new game, 0-5 = load savegame
	short   episode;                // 1-3
	short   map;                    // 1-9
	short   skill;                  // 1-5

// info specific to this node
	short   consoleplayer;
	short   numplayers;
	short   angleoffset;    // 1 = left, 0 = center, -1 = right
	short   drone;                  // 1 = drone

// packet data to be sent
	doomdata_t      data;
} doomcom_t;
*/ // FUCKED LINES

extern  doomcom_t               *doomcom;

/*
====================
=
= I_InitNetwork
=
====================
*/

void I_InitNetwork (void)
{
	int             i;

	i = M_CheckParm ("-net");
	if (!i)
	{
	//
	// single player game
	//
		doomcom = malloc (sizeof (*doomcom) );
		memset (doomcom, 0, sizeof(*doomcom) );
		netgame = false;
		doomcom->id = DOOMCOM_ID;
		doomcom->numplayers = doomcom->numnodes = 1;
		doomcom->deathmatch = false;
		doomcom->consoleplayer = 0;
		doomcom->ticdup = 1;
		doomcom->extratics = 0;
		return;
	}

	netgame = true;
	doomcom = (doomcom_t *)atoi(myargv[i+1]);
//DEBUG
doomcom->skill = startskill;
doomcom->episode = startepisode;
doomcom->map = startmap;
doomcom->deathmatch = deathmatch;

}

void I_NetCmd (void)
{
	// NET_TODO:
/*
	if (!netgame)
		I_Error ("I_NetCmd when not in netgame");
	DPMIInt (doomcom->intnum);
*/
}

//=========================================================================
//=========================================================================
// Hi-Res (mode 12) stuff
//=========================================================================
//=========================================================================

HRESULT WINAPI EnumModesCallback2(DDSURFACEDESC2 *pDDSD, void *pvData)
{
	if( (pDDSD->dwWidth <= 1600)	&& 
		(pDDSD->dwHeight <= 1200)	&&
		(!(pDDSD->ddsCaps.dwCaps & DDSCAPS_MODEX)) )
	{
		if(g_dwcResolutions)
		{
			g_rgScreenWidth = (DWORD *)realloc(g_rgScreenWidth, (g_dwcResolutions + 1) * sizeof(DWORD));
			g_rgScreenHeight = (DWORD *)realloc(g_rgScreenHeight, (g_dwcResolutions + 1) * sizeof(DWORD));
		}
		else
		{
			g_rgScreenWidth = (DWORD *)malloc(sizeof(DWORD));
			g_rgScreenHeight = (DWORD *)malloc(sizeof(DWORD));
		}

		g_rgScreenWidth[g_dwcResolutions] = pDDSD->dwWidth;
		g_rgScreenHeight[g_dwcResolutions] = pDDSD->dwHeight;
		g_dwcResolutions += 1;
	}

	return DDENUMRET_OK;
}


//==========================================================================
//
// SetVideoModeHR - Set video mode to 640x480x16
//
//==========================================================================


void SetVideoModeHR(void)
{
	HRESULT hr;
	IDirectDraw *pDD = NULL;
	DDSURFACEDESC2 ddsd;
	HDC hdc;
	int i;

	if(novideo)
	{
		return;
	}

	// Ok, now setup DirectDraw...
	//
	if(FAILED(hr = DirectDrawCreate(NULL, &pDD, NULL)))
	{
		I_Error("Could not initialize DirectDraw! (Error code 0x%08x)", hr);
	}

	hr = IDirectDraw_QueryInterface(pDD, &IID_IDirectDraw4, (void **)&g_pDD);
	IDirectDraw_Release(pDD);

	if(FAILED(hr))
	{
		I_Error("Could not initialize DirectDraw! (Error code 0x%08x)", hr);
	}

	// Enumerate all availble video modes for the menu later...
	//
	ZeroMemory(&ddsd, sizeof(DDSURFACEDESC2));

	ddsd.dwSize = sizeof(DDSURFACEDESC2);
	ddsd.dwFlags = DDSD_PIXELFORMAT;
	ddsd.ddpfPixelFormat.dwSize = sizeof(DDPIXELFORMAT);
	ddsd.ddpfPixelFormat.dwFlags = DDPF_PALETTEINDEXED8 | DDPF_RGB;
	ddsd.ddpfPixelFormat.dwRGBBitCount = 8;

	if(FAILED(hr = IDirectDraw4_EnumDisplayModes(g_pDD, DDEDM_STANDARDVGAMODES, &ddsd, NULL, EnumModesCallback2)))
	{
		I_Error("Could not retrieve list of available video modes! (Error code 0x%08x)", hr);
	}

	if(FAILED(hr = IDirectDraw4_SetCooperativeLevel(g_pDD, g_hWnd, DDSCL_EXCLUSIVE | DDSCL_FULLSCREEN | DDSCL_ALLOWREBOOT)))
	{
		I_Error("Could not SetCooperativeLevel! (Error code 0x%08x)", hr);
	}

	if(FAILED(hr = IDirectDraw4_SetDisplayMode(g_pDD, 640, 480, 8, 0, 0)))
	{
		I_Error("Could not set a fullscreen resolution of %dx%d (Error code 0x%08x)", g_ScreenWidth, g_ScreenHeight, hr);
	}

	ZeroMemory(&ddsd, sizeof(DDSURFACEDESC2));

	ddsd.dwSize = sizeof(DDSURFACEDESC2);
	ddsd.dwFlags = DDSD_CAPS;
//	ddsd.dwWidth = SCREENWIDTH;
//	ddsd.dwHeight = SCREENHEIGHT;
	ddsd.ddsCaps.dwCaps = DDSCAPS_PRIMARYSURFACE;

	if(FAILED(hr = IDirectDraw4_CreateSurface(g_pDD, &ddsd, &g_pDDS, NULL)))
	{
		I_Error("Could not create DirectDraw Surface for startup screen! (Error code 0x%08x)", hr);
	}

	hdc = GetDC(NULL);

	GetSystemPaletteEntries(hdc, 0, 256, g_palOriginal);
	SetSystemPaletteUse(hdc, SYSPAL_NOSTATIC);

	ReleaseDC(NULL, hdc);

	memcpy(g_palLastSet, g_palOriginal, 256 * sizeof(PALETTEENTRY));

	for (i = 0; i < 256; i++) 
		g_palLastSet[i].peFlags = PC_NOCOLLAPSE | PC_RESERVED;

//	g_palLastSet[0].peFlags = PC_EXPLICIT;
//	g_palLastSet[0].peRed = 0;
//	g_palLastSet[255].peFlags = PC_EXPLICIT;
//	g_palLastSet[255].peRed = 255;

	if(FAILED(hr = IDirectDraw4_CreatePalette(g_pDD, 
											  DDPCAPS_8BIT | DDPCAPS_ALLOW256,
											  g_palLastSet, 
											  &g_pDDP, 
											  NULL)))
	{
		I_Error("Could not create DirectDraw Palette! (Error code 0x%08x)", hr);
	}

	if(FAILED(hr = IDirectDrawSurface4_SetPalette(g_pDDS, g_pDDP)))
	{
		I_Error("Could not set Palette! (Error code 0x%08x)", hr);
	}


/*
	union REGS regs;
	regs.x.eax = 0x12;
	int386(VID_INT, &regs, &regs);
*/
}


void ShutdownHR(void)
{
	// We really only need to destroy the surface we created. The other DirectDraw
	// objects are all ok to use later...
	//

	if(g_pDDS)
	{
		IDirectDrawSurface4_SetPalette(g_pDDS, NULL);
		IDirectDrawSurface4_SetClipper(g_pDDS, NULL);
		IDirectDrawSurface4_Release(g_pDDS);
		g_pDDS = NULL;
	}
}


void GetPaletteHR(byte *palette);
void WritePCXfile (char *filename, byte *data, int width, int height, byte *palette);

//==========================================================================
//
// ClearScreenHR - Clear the screen to color 0
//
//==========================================================================

void ClearScreenHR(void)
{
	DDSURFACEDESC2 ddsd;
	HRESULT hr;
//	char pal[768];

	if(!g_pDDS)
	{
		return;
	}

//	GetPaletteHR(pal);

	ZeroMemory(&ddsd, sizeof(DDSURFACEDESC2));

	ddsd.dwSize = sizeof(DDSURFACEDESC2);

	if(FAILED(hr = IDirectDrawSurface4_Lock(g_pDDS, NULL, &ddsd, DDLOCK_SURFACEMEMORYPTR, NULL)))
	{
		I_Error("Could not lock DirectDraw Surface! (Error code 0x%08x)", hr);
	}

//	WritePCXfile("Startup.pcx", ddsd.lpSurface, 640, 480, pal);

	ZeroMemory(ddsd.lpSurface, 640 * 480);

	hr = IDirectDrawSurface4_Unlock(g_pDDS, NULL);

/*
	BITPLANE(MASK_PLANE0|MASK_PLANE1|MASK_PLANE2|MASK_PLANE3);
	memset((char *)0xa0000,0,38400);
*/
}


//==========================================================================
//
// SlamHR - copy 4-plane buffer to screen
//
//==========================================================================

void SlamHR(char *buffer)
{
	HRESULT hr;
	int x, y, xChunk;
	DDSURFACEDESC2 ddsd;
	byte *pR, *pG, *pB, *pI;
	byte *pSurface;

	if(!g_pDDS)
	{
		return;
	}

	ZeroMemory(&ddsd, sizeof(DDSURFACEDESC2));
	ddsd.dwSize = sizeof(DDSURFACEDESC2);

	if(FAILED(hr = IDirectDrawSurface4_Lock(g_pDDS, NULL, &ddsd, DDLOCK_SURFACEMEMORYPTR | DDLOCK_WRITEONLY, NULL)))
	{
		I_Error("Could not lock DirectDraw Surface! (Error code 0x%08x)", hr);
	}

	// The buffer passed in is in native 640x480x4 format
	// (i.e. each bit plane is in discrete memory locations.)
	//
	// We'll merge it on the fly to 8-bit indexes (0 - 15) 
	// as we dump it to VRAM.
	//
	pB = buffer + P0OFFSET;
	pG = buffer + P1OFFSET;
	pR = buffer + P2OFFSET;
	pI = buffer + P3OFFSET;

	for(y = 0; y < 480; y++)
	{
		pSurface = ((byte *)ddsd.lpSurface) + (y * ddsd.lPitch);

		for(xChunk = 0; xChunk < 80; xChunk++, pSurface += 8, pR++, pG++, pB++, pI++)
		{
			byte pixBuf[8];

			for(x = 0; x < 8; x++)
			{
				int xShift = 7 - x;

				pixBuf[x] = ((*pB) >> xShift) & 0x01;
				pixBuf[x] |= (((*pG) >> xShift) & 0x01) << 1;
				pixBuf[x] |= (((*pR) >> xShift) & 0x01) << 2;
				pixBuf[x] |= (((*pI) >> xShift) & 0x01) << 3;
			}

			pSurface[0] = pixBuf[0];
			pSurface[1] = pixBuf[1];
			pSurface[2] = pixBuf[2];
			pSurface[3] = pixBuf[3];
			pSurface[4] = pixBuf[4];
			pSurface[5] = pixBuf[5];
			pSurface[6] = pixBuf[6];
			pSurface[7] = pixBuf[7];
		}
	}

	hr = IDirectDrawSurface4_Unlock(g_pDDS, NULL);

/*
	BITPLANE(MASK_PLANE0);
	memcpy((char *)0xA0000, buffer+P0OFFSET, 38400);
	BITPLANE(MASK_PLANE1);
	memcpy((char *)0xA0000, buffer+P1OFFSET, 38400);
	BITPLANE(MASK_PLANE2);
	memcpy((char *)0xA0000, buffer+P2OFFSET, 38400);
	BITPLANE(MASK_PLANE3);
	memcpy((char *)0xA0000, buffer+P3OFFSET, 38400);
*/
}


//==========================================================================
//
// SlamHR - copy 4-plane buffer to screen
//
// X and Width should be a multiple of 8
// src should be 4 planes of block size, back to back
//==========================================================================

void SlamBlockHR(int x, int y, int w, int h, char *src)
{
	HRESULT hr;
	int startX, startY, xChunk;
	DDSURFACEDESC2 ddsd;
	byte *pR, *pG, *pB, *pI;
	byte *pSurface;
	int bytesPerBlock;

	if(!g_pDDS)
	{
		return;
	}

	ZeroMemory(&ddsd, sizeof(DDSURFACEDESC2));
	ddsd.dwSize = sizeof(DDSURFACEDESC2);

	if(FAILED(hr = IDirectDrawSurface4_Lock(g_pDDS, NULL, &ddsd, DDLOCK_SURFACEMEMORYPTR | DDLOCK_WRITEONLY, NULL)))
	{
		I_Error("Could not lock DirectDraw Surface! (Error code 0x%08x)", hr);
	}

	bytesPerBlock = (w / 8) * h;

	// The buffer passed in is in native 640x480x4 format
	// (i.e. each bit plane is in discrete memory locations.)
	//
	// We'll merge it on the fly to 8-bit indexes (0 - 15) 
	// as we dump it to VRAM.
	//
	pB = src;
	pG = src + bytesPerBlock;
	pR = src + (bytesPerBlock * 2);
	pI = src + (bytesPerBlock * 3);

	for(startY = y; startY < (y + h); startY++)
	{
		pSurface = ((byte *)ddsd.lpSurface) + (startY * ddsd.lPitch) + x;

		for(xChunk = 0; xChunk < (w / 8); xChunk++, pSurface += 8, pR++, pG++, pB++, pI++)
		{
			byte pixBuf[8];

			for(startX = 0; startX < 8; startX++)
			{
				int xShift = 7 - startX;

				pixBuf[startX] = ((*pB) >> xShift) & 0x01;
				pixBuf[startX] |= (((*pG) >> xShift) & 0x01) << 1;
				pixBuf[startX] |= (((*pR) >> xShift) & 0x01) << 2;
				pixBuf[startX] |= (((*pI) >> xShift) & 0x01) << 3;
			}

			pSurface[0] = pixBuf[0];
			pSurface[1] = pixBuf[1];
			pSurface[2] = pixBuf[2];
			pSurface[3] = pixBuf[3];
			pSurface[4] = pixBuf[4];
			pSurface[5] = pixBuf[5];
			pSurface[6] = pixBuf[6];
			pSurface[7] = pixBuf[7];
		}
	}

	hr = IDirectDrawSurface4_Unlock(g_pDDS, NULL);

/*
	int srcwid = w>>3;
	char *dest = ((char *)0xA0000) + (y*(640/8)) + (x>>3);
	char *dst;
	int i;

	VB_SYNC;

	BITPLANE(MASK_PLANE0);
	dst = dest;
	for ( i=0; i<h; i++ )
	{
		memcpy(dst, src, srcwid);
		dst += 640/8;
		src += srcwid;
	}
	BITPLANE(MASK_PLANE1);
	dst = dest;
	for ( i=0; i<h; i++ )
	{
		memcpy(dst, src, srcwid);
		dst += 640/8;
		src += srcwid;
	}
	BITPLANE(MASK_PLANE2);
	dst = dest;
	for ( i=0; i<h; i++ )
	{
		memcpy(dst, src, srcwid);
		dst += 640/8;
		src += srcwid;
	}
	BITPLANE(MASK_PLANE3);
	dst = dest;
	for ( i=0; i<h; i++ )
	{
		memcpy(dst, src, srcwid);
		dst += 640/8;
		src += srcwid;
	}
*/
}

//==========================================================================
//
// InitPaletteHR
//
//==========================================================================

void InitPaletteHR(void)
{
/*
	int i;
	union REGS regs;

	// Set palette registers to point into color registers
	for ( i=0; i<16; i++ )
	{
		regs.x.eax = (0x10<<8)|0x00;
		regs.x.ebx = (i<<8)|i;
		int386(VID_INT, &regs, &regs);
	}
*/
}


//==========================================================================
//
// SetPaletteHR - Set the HR palette
//
//==========================================================================

void SetPaletteHR(byte *palette)
{
	int i;
	HRESULT hr;

	if(!g_pDDP || !g_pDD)
	{
		return;
	}

	for(i = 0; i < 16; i++, palette += 3)
	{
		g_palLastSet[i].peRed = palette[0] ? (palette[0] << 2) | 0x03 : 0;
		g_palLastSet[i].peGreen = palette[1] ? (palette[1] << 2) | 0x03 : 0;
		g_palLastSet[i].peBlue = palette[2] ? (palette[2] << 2) | 0x03 : 0;
	}

	hr = IDirectDraw4_WaitForVerticalBlank(g_pDD, DDWAITVB_BLOCKBEGIN, NULL);
	hr = IDirectDrawPalette_SetEntries(g_pDDP, 0, 0, 16, g_palLastSet);

/*
	int i;
	VB_SYNC;
	outp(PEL_WRITE_ADR, 0);

	for(i = 0; i < 16*3; i++)
	{
		outp(PEL_DATA, (*palette++));
	}
*/
}


//==========================================================================
//
// GetPaletteHR - Get the HR palette
//
//==========================================================================

void GetPaletteHR(byte *palette)
{
	PALETTEENTRY curPal[16];
	int i;
	HRESULT hr;

	if(!g_pDDP)
	{
		return;
	}

	hr = IDirectDrawPalette_GetEntries(g_pDDP, 0, 0, 16, curPal);

	for(i = 0; i < 16; i++, palette += 3)
	{
		palette[0] = curPal[i].peRed;
		palette[1] = curPal[i].peGreen;
		palette[2] = curPal[i].peBlue;
	}

/*
	int i;
	outp(PEL_READ_ADR, 0);
	for (i=0; i<16*3; i++)
	{
		*palette++ = inp(PEL_DATA);
	}
*/
}


//==========================================================================
//
// FadeToPaletteHR
//
//==========================================================================

void FadeToPaletteHR(byte *palette)
{
	int i,j;
	int steps=70;          // two-seconds
//	int steps=140;          // two-seconds
	byte basep[16*3];
	byte work[16*3];
	int delta;

	GetPaletteHR(basep);
	for(i = 0; i < steps; i++)
	{
		for(j = 0; j < 16*3; j++)
		{
			delta = palette[j]-basep[j];
			work[j] = basep[j]+delta*i/steps;
		}
		SetPaletteHR(work);
	}
	SetPaletteHR(palette);
}


//==========================================================================
//
// FadeToBlackHR - Fades the palette out to black
//
//==========================================================================


void FadeToBlackHR(void)
{
	char work[16*3];
	char base[16*3];
	int i,j;
	int steps=105;
//	int steps=70;

	GetPaletteHR(base);
	for (i=0; i<steps; i++)
	{
		for (j=0; j<16*3; j++)
		{
			work[j] = base[j]-(base[j]*i/steps);
		}
//		VB_SYNC;
		SetPaletteHR(work);
	}
	memset(work,0,16*3);
	SetPaletteHR(work);
}


//==========================================================================
//
// BlackPaletteHR - Instantly blacks out the palette
//
//==========================================================================

void BlackPaletteHR(void)
{
	char blackpal[16*3];

	memset(blackpal,0,16*3);
	SetPaletteHR(blackpal);
}


//==========================================================================
//
//
// I_StartupReadKeys
//
//
//==========================================================================

void I_StartupReadKeys(void)
{
/*
	int k;

	while (kbdtail < kbdhead)
	{
	   k = keyboardque[kbdtail&(KBDQUESIZE-1)];
	   kbdtail++;
	   if (k == 1)
		   I_Quit ();
	}
*/
}
