
//**************************************************************************
//**
//** r_plane.c : Heretic 2 : Raven Software, Corp.
//**
//** $RCSfile: r_plane.c,v $
//** $Revision: 1.5 $
//** $Date: 95/07/13 15:17:12 $
//** $Author: cjr $
//**
//**************************************************************************

// HEADER FILES ------------------------------------------------------------

#include "h2def.h"
#include "r_local.h"
#include <crtdbg.h>

// MACROS ------------------------------------------------------------------

// TYPES -------------------------------------------------------------------

// EXTERNAL FUNCTION PROTOTYPES --------------------------------------------

// PUBLIC FUNCTION PROTOTYPES ----------------------------------------------

// PRIVATE FUNCTION PROTOTYPES ---------------------------------------------

// EXTERNAL DATA DECLARATIONS ----------------------------------------------

extern fixed_t Sky1ScrollDelta;
extern fixed_t Sky2ScrollDelta;

// PUBLIC DATA DEFINITIONS -------------------------------------------------

int Sky1Texture;
int Sky2Texture;
fixed_t Sky1ColumnOffset;
fixed_t Sky2ColumnOffset;
int skyflatnum;
int skytexturemid;
fixed_t skyiscale;
boolean DoubleSky;
planefunction_t floorfunc, ceilingfunc;

// Opening
visplane_t visplanes[MAXVISPLANES], *lastvisplane;
visplane_t *floorplane, *ceilingplane;
short openings[MAXOPENINGS], *lastopening;

// Clip values are the solid pixel bounding the range.
// floorclip start out SCREENHEIGHT
// ceilingclip starts out -1
short floorclip[MAXWIDTH];
short ceilingclip[MAXWIDTH];

// spanstart holds the start of a plane span, initialized to 0
int spanstart[MAXHEIGHT];
int spanstop[MAXHEIGHT];

// Texture mapping
lighttable_t **planezlight;
fixed_t planeheight;
fixed_t yslope[MAXHEIGHT];
fixed_t distscale[MAXWIDTH];
fixed_t basexscale, baseyscale;
fixed_t cachedheight[MAXHEIGHT];
fixed_t cacheddistance[MAXHEIGHT];
fixed_t cachedxstep[MAXHEIGHT];
fixed_t cachedystep[MAXHEIGHT];

// PRIVATE DATA DEFINITIONS ------------------------------------------------

static byte *g_rgSky1Buffer = NULL;
static byte *g_rgSky2Buffer = NULL;
static int g_ActualSky1Texture;
static int g_ActualSky2Texture;

// SkyHeight will always be equal to g_ScreenHeight...
//
static int g_cbSkyWidth = 0;
static int g_AngleToSkyShift;

// CODE --------------------------------------------------------------------


//==========================================================================
//
// R_InitSky
//
// Called at level load.
//
//==========================================================================

extern short **texturecolumnlump;
extern unsigned short		**texturecolumnofs;

void R_InitSky(int map)
{
	int skyTexture, skyTexture2;
	byte skyBuffer[256 * 200];
	fixed_t hScale, wScale;
	int i;
	int nTotalSkyAngles;
	__int64 lltemp;

	// Only the following chunk is from the original code...
	//
	Sky1Texture = P_GetMapSky1Texture(map);
	Sky2Texture = P_GetMapSky2Texture(map);
	Sky1ScrollDelta = P_GetMapSky1ScrollDelta(map);
	Sky2ScrollDelta = P_GetMapSky2ScrollDelta(map);
	Sky1ColumnOffset = 0;
	Sky2ColumnOffset = 0;
	DoubleSky = P_GetMapDoubleSky(map);

	// The rest of this is added by me...
	//
	if(g_rgSky1Buffer)
	{
		free(g_rgSky1Buffer);
	}

	if(g_rgSky2Buffer)
	{
		free(g_rgSky2Buffer);
	}

	hScale = FixedDiv(g_ScreenHeight * FRACUNIT, 200 * FRACUNIT);
	g_cbSkyWidth = FixedMul(256 * FRACUNIT, hScale);
	wScale = FixedDiv(g_cbSkyWidth, 256 * FRACUNIT);

	Sky1ScrollDelta = FixedMul(Sky1ScrollDelta, wScale);
	Sky2ScrollDelta = FixedMul(Sky2ScrollDelta, wScale);

	// Now convert g_cbSkyWidth from fixed-point to regular int.
	//
	g_cbSkyWidth >>= FRACBITS;

	nTotalSkyAngles = g_cbSkyWidth * 4;
	lltemp = 0x100000000i64;
	g_AngleToSkyShift = lltemp / nTotalSkyAngles;

	g_rgSky1Buffer = malloc(g_cbSkyWidth * g_ScreenHeight);
	g_rgSky2Buffer = malloc(g_cbSkyWidth * g_ScreenHeight);

	skyTexture = texturetranslation[Sky1Texture];
	skyTexture2 = texturetranslation[Sky2Texture];
	g_ActualSky1Texture = Sky1Texture;
	g_ActualSky2Texture = Sky2Texture;

	for(i = 0; i < 256; i++)
	{
		int lump = texturecolumnlump[skyTexture][i];
		int ofs = texturecolumnofs[skyTexture][i];
		byte *pCol = (byte *)W_CacheLumpNum(lump,PU_CACHE) + ofs;

		memcpy(skyBuffer + (i * 200), pCol, 200);
	}

	// Now, create DirectDraw surfaces out of our "scaled-size" memory
	// and our "original-size" memory...
	//
	I_DDStretchBlt(g_rgSky1Buffer, g_ScreenHeight, g_cbSkyWidth, skyBuffer, 200, 256);

	for(i = 0; i < 256; i++)
	{
		int lump = texturecolumnlump[skyTexture2][i];
		int ofs = texturecolumnofs[skyTexture2][i];
		byte *pCol = (byte *)W_CacheLumpNum(lump,PU_CACHE) + ofs;

		memcpy(skyBuffer + (i * 200), pCol, 200);
	}

	I_DDStretchBlt(g_rgSky2Buffer, g_ScreenHeight, g_cbSkyWidth, skyBuffer, 200, 256);

	// Okay, we posess scaled versions of both skies now.
	// (i.e. we're done here...)
	//
}

//==========================================================================
//
// R_InitSkyMap
//
// Called whenever the view size changes.
//
//==========================================================================

void R_InitSkyMap(void)
{
	skyflatnum = R_FlatNumForName("F_SKY");
	skytexturemid = 200*FRACUNIT;
	skyiscale = FRACUNIT;
}

//==========================================================================
//
// R_InitPlanes
//
// Called at game startup.
//
//==========================================================================

void R_InitPlanes(void)
{
}

//==========================================================================
//
// R_MapPlane
//
// Globals used: planeheight, ds_source, basexscale, baseyscale,
// viewx, viewy.
//
//==========================================================================

void R_MapPlane(int y, int x1, int x2)
{
	angle_t angle;
	fixed_t distance, length;
	unsigned index;

#ifdef RANGECHECK
	if(x2 < x1 || x1 < 0 || x2 >= viewwidth || (unsigned)y > viewheight)
	{
		I_Error("R_MapPlane: %i, %i at %i (when viewwidth=%d and viewheight=%d)", x1, x2, y, viewwidth, viewheight);
	}
#endif

	if(planeheight != cachedheight[y])
	{
		cachedheight[y] = planeheight;
		distance = cacheddistance[y] = FixedMul(planeheight, yslope[y]);
		ds_xstep = cachedxstep[y] = FixedMul(distance, basexscale);
		ds_ystep = cachedystep[y] = FixedMul(distance, baseyscale);
	}
	else
	{
		distance = cacheddistance[y];
		ds_xstep = cachedxstep[y];
		ds_ystep = cachedystep[y];
	}
	
	length = FixedMul(distance, distscale[x1]);
	angle = (viewangle+xtoviewangle[x1])>>ANGLETOFINESHIFT;
	ds_xfrac = viewx+FixedMul(finecosine[angle], length);
	ds_yfrac = -viewy-FixedMul(finesine[angle], length);

	if(fixedcolormap)
	{
		ds_colormap = fixedcolormap;
	}
	else
	{
		index = distance >> LIGHTZSHIFT;
		if(index >= MAXLIGHTZ )
		{
			index = MAXLIGHTZ-1;
		}
		ds_colormap = planezlight[index];
	}

	ds_y = y;
	ds_x1 = x1;
	ds_x2 = x2;

	spanfunc(); // High or low detail
}

//==========================================================================
//
// R_ClearPlanes
//
// Called at the beginning of each frame.
//
//==========================================================================

void R_ClearPlanes(void)
{
	int i;
	angle_t angle;

	// Opening / clipping determination
	for(i = 0; i < viewwidth; i++)
	{
		floorclip[i] = viewheight;
		ceilingclip[i] = -1;
	}

	lastvisplane = visplanes;
	lastopening = openings;

	// Texture calculation
	memset(cachedheight, 0, sizeof(cachedheight));	
	angle = (viewangle-ANG90)>>ANGLETOFINESHIFT; // left to right mapping
	// Scale will be unit scale at SCREENWIDTH/2 distance
	basexscale = FixedDiv(finecosine[angle], centerxfrac);
	baseyscale = -FixedDiv(finesine[angle], centerxfrac);
}

//==========================================================================
//
// R_FindPlane
//
//==========================================================================

visplane_t *R_FindPlane(fixed_t height, int picnum,
	int lightlevel, int special)
{
	visplane_t *check;

	if(special < 150)
	{ // Don't let low specials affect search
		special = 0;
	}

	if(picnum == skyflatnum)
	{ // All skies map together
		height = 0;
		lightlevel = 0;
	}

	for(check = visplanes; check < lastvisplane; check++)
	{
		if(height == check->height
		&& picnum == check->picnum
		&& lightlevel == check->lightlevel
		&& special == check->special)
			break;
	}

	if(check < lastvisplane)
	{
		return(check);
	}

	if(lastvisplane-visplanes == MAXVISPLANES)
	{
		I_Error("R_FindPlane: no more visplanes");
	}

	lastvisplane++;
	check->height = height;
	check->picnum = picnum;
	check->lightlevel = lightlevel;
	check->special = special;
	check->minx = g_ScreenWidth;
	check->maxx = -1;
	memset(check->top, 0xff, sizeof(check->top));
	return(check);
}

//==========================================================================
//
// R_CheckPlane
//
//==========================================================================

visplane_t *R_CheckPlane(visplane_t *pl, int start, int stop)
{
	int intrl, intrh;
	int unionl, unionh;
	int x;

	if(start < pl->minx)
	{
		intrl = pl->minx;
		unionl = start;
	}
	else
	{
		unionl = pl->minx;
		intrl = start;
	}
	if(stop > pl->maxx)
	{
		intrh = pl->maxx;
		unionh = stop;
	}
	else
	{
		unionh = pl->maxx;
		intrh = stop;
	}

	for(x = intrl; x <= intrh; x++)
	{
		if(pl->top[x] != 0xffff)
		{
			break;
		}
	}

	if(x > intrh)
	{
		pl->minx = unionl;
		pl->maxx = unionh;
		return pl; // use the same visplane
	}

	// Make a new visplane
	lastvisplane->height = pl->height;
	lastvisplane->picnum = pl->picnum;
	lastvisplane->lightlevel = pl->lightlevel;
	lastvisplane->special = pl->special;
	pl = lastvisplane++;
	pl->minx = start;
	pl->maxx = stop;
	memset(pl->top, 0xff, sizeof(pl->top));

	return pl;
}

//==========================================================================
//
// R_MakeSpans
//
//==========================================================================

void R_MakeSpans(int x, int t1, int b1, int t2, int b2)
{
	while(t1 < t2 && t1 <= b1)
	{
		R_MapPlane(t1, spanstart[t1], x-1);
		t1++;
	}
	while(b1 > b2 && b1 >= t1)
	{
		R_MapPlane(b1, spanstart[b1], x-1);
		b1--;
	}
	while(t2 < t1 && t2 <= b2)
	{
		spanstart[t2] = x;
		t2++;
	}
	while(b2 > b1 && b2 >= t2)
	{
		spanstart[b2] = x;
		b2--;
	}
}

//==========================================================================
//
// R_DrawPlanes
//
//==========================================================================

//#define SKYTEXTUREMIDSHIFTED SCREENHEIGHT
#define SKYTEXTUREMIDSHIFTED 200

void R_DrawPlanes(void)
{
	visplane_t *pl;
	int light;
	int x, stop;
	int angle;
	byte *tempSource;
	byte *source;
	byte *source2;
	byte *dest;
	int count;
	int offset;
	int skyTexture;
	int offset2;
	int skyTexture2;
	int scrollOffset;

	extern byte *ylookup[MAXHEIGHT];
	extern int columnofs[MAXWIDTH];

#ifdef RANGECHECK
	if(ds_p-drawsegs > MAXDRAWSEGS)
	{
		I_Error("R_DrawPlanes: drawsegs overflow (%i)", ds_p-drawsegs);
	}
	if(lastvisplane-visplanes > MAXVISPLANES)
	{
		I_Error("R_DrawPlanes: visplane overflow (%i)",
			lastvisplane-visplanes);
	}
	if(lastopening-openings > MAXOPENINGS)
	{
		I_Error("R_DrawPlanes: opening overflow (%i)",
			lastopening-openings);
	}
#endif

	for(pl = visplanes; pl < lastvisplane; pl++)
	{
		if(pl->minx > pl->maxx)
		{
			continue;
		}
		if(pl->picnum == skyflatnum)
		{ // Sky flat
			if(DoubleSky)
			{ 
				// Render 2 layers, sky 1 in front
				//
				offset = Sky1ColumnOffset>>16;
				offset2 = Sky2ColumnOffset>>16;
				for(x = pl->minx; x <= pl->maxx; x++)
				{
					dc_yl = pl->top[x];
					dc_yh = pl->bottom[x];
					if(dc_yl <= dc_yh)
					{
						count = dc_yh-dc_yl;
						if(count < 0)
						{
							return;
						}

						angle = (viewangle + xtoviewangle[x]) / g_AngleToSkyShift;	// was >> ANGLETOSKYSHIFT

						source = g_rgSky1Buffer + 
								 (((angle + offset) % g_cbSkyWidth) * g_ScreenHeight) +
								 g_ScreenHeight +	// was SKYTEXTUREMIDSHIFTED
								 (dc_yl - centery);

						source2 = g_rgSky2Buffer + 
								  (((angle + offset2) % g_cbSkyWidth) * g_ScreenHeight) +
								  g_ScreenHeight +	// was SKYTEXTUREMIDSHIFTED
								  (dc_yl - centery);

						dest = ylookup[dc_yl]+columnofs[x];
						do
						{
							if(*source)
							{
								*dest = *source++;
								source2++;
							}
							else
							{
								*dest = *source2++;
								source++;
							}
							dest += g_ScreenWidth;
						} while(count--);

// Old "mirror" code...
/* 
				offset = Sky1ColumnOffset>>16;
				skyTexture = texturetranslation[Sky1Texture];
				offset2 = Sky2ColumnOffset>>16;
				skyTexture2 = texturetranslation[Sky2Texture];
				for(x = pl->minx; x <= pl->maxx; x++)
				{
					dc_yl = pl->top[x];
					dc_yh = pl->bottom[x];
					if(dc_yl <= dc_yh)
					{
						int y;
						int quot, rem;
						int srcOff;

						angle = (viewangle+xtoviewangle[x])
								>>ANGLETOSKYSHIFT;
						source = R_GetColumn(skyTexture, angle+offset);
						source2 = R_GetColumn(skyTexture2, angle + offset2);

						srcOff = g_ScreenHeight + 100 + (dc_yl - centery);
						// This should never be a negative number!
						//
						_ASSERTE(srcOff >= 0);

						dest = ylookup[dc_yl]+columnofs[x];
						quot = srcOff / 200;
						rem = srcOff % 200;

						for(y = dc_yl; y <= dc_yh; y++, dest += g_ScreenWidth, rem++)
						{
							byte bSrc, bSrc2;

							// If the quotient is EVEN (0,2,4,etc.) we'll count from
							// to beginning towards the end.
							// If the quotient is ODD (1,3,5,etc.) we'll count from
							// the end towards the beginning.
							//
							if(rem >= 200)
							{
								rem = 0;
								quot += 1;
							}

							if(quot & 0x01)
							{
								// ODD!
								//
								int mirror = 199 - rem;

								bSrc = *(source + mirror);
								bSrc2 = *(source2 + mirror);
							}
							else
							{
								// EVEN!
								//
								bSrc = *(source + rem);
								bSrc2 = *(source2 + rem);
							}
							if(bSrc)
							{
								*dest = bSrc;
							}
							else
							{
								*dest = bSrc2;
							}
						}
*/
/*
						count = dc_yh-dc_yl;
						if(count < 0)
						{
							return;
						}
						angle = (viewangle+xtoviewangle[x])
							>>ANGLETOSKYSHIFT;
						source = R_GetColumn(skyTexture, angle+offset)
							+SKYTEXTUREMIDSHIFTED+(dc_yl-centery);
						source2 = R_GetColumn(skyTexture2, angle+offset2)
							+SKYTEXTUREMIDSHIFTED+(dc_yl-centery);
						dest = ylookup[dc_yl]+columnofs[x];
						do
						{
							if(*source)
							{
								*dest = *source++;
								source2++;
							}
							else
							{
								*dest = *source2++;
								source++;
							}
							dest += SCREENWIDTH;
						} while(count--);
*/
					}
				}
				continue; // Next visplane
			}
			else
			{ 
				byte *skyBuffer;

				// Render single layer
				//
				if(pl->special == 200)
				{ // Use sky 2
					offset = Sky2ColumnOffset>>16;
					skyTexture = texturetranslation[Sky2Texture];
					if(Sky2Texture == g_ActualSky2Texture)
					{
						skyBuffer = g_rgSky2Buffer;
					}
					else
					{
						skyBuffer = g_rgSky1Buffer;
					}
				}
				else
				{ // Use sky 1
					offset = Sky1ColumnOffset>>16;
					skyTexture = texturetranslation[Sky1Texture];

					// Here we account for lightning, etc., when
					// other parts of the program change what
					// 'Sky1Texture' really is (without telling us!)
					//
					if(Sky1Texture == g_ActualSky1Texture)
					{
						skyBuffer = g_rgSky1Buffer;
					}
					else
					{
						skyBuffer = g_rgSky2Buffer;
					}
				}
				for(x = pl->minx; x <= pl->maxx; x++)
				{
					dc_yl = pl->top[x];
					dc_yh = pl->bottom[x];
					if(dc_yl <= dc_yh)
					{
						count = dc_yh-dc_yl;
						if(count < 0)
						{
							return;
						}

						angle = (viewangle + xtoviewangle[x]) / g_AngleToSkyShift;	// was >> ANGLETOSKYSHIFT

						source = skyBuffer + 
								 (((angle + offset) % g_cbSkyWidth) * g_ScreenHeight) +
								 g_ScreenHeight +	// was SKYTEXTUREMIDSHIFTED
								 (dc_yl - centery);

						dest = ylookup[dc_yl]+columnofs[x];
						do
						{
							*dest = *source++;
							dest += g_ScreenWidth;
						} while(count--);

// Old "mirror" code...
/* 
						int y;
						int quot, rem;
						int srcOff;

						angle = (viewangle+xtoviewangle[x])
								>>ANGLETOSKYSHIFT;
						source = R_GetColumn(skyTexture, angle+offset);

						srcOff = g_ScreenHeight + 100 + (dc_yl - centery);
						// This should never be a negative number!
						//
						_ASSERTE(srcOff >= 0);

						dest = ylookup[dc_yl]+columnofs[x];
						quot = srcOff / 200;
						rem = srcOff % 200;

						for(y = dc_yl; y <= dc_yh; y++, dest += g_ScreenWidth, rem++)
						{
							// If the quotient is EVEN (0,2,4,etc.) we'll count from
							// to beginning towards the end.
							// If the quotient is ODD (1,3,5,etc.) we'll count from
							// the end towards the beginning.
							//
							if(rem >= 200)
							{
								rem = 0;
								quot += 1;
							}

							if(quot & 0x01)
							{
								// ODD!
								//
								*dest = *(source + (199 - rem));
							}
							else
							{
								// EVEN!
								//
								*dest = *(source + rem);
							}
						}
*/
/*
						count = dc_yh-dc_yl;
						if(count < 0)
						{
							return;
						}
						source = R_GetColumn(skyTexture, angle+offset)
							     +SKYTEXTUREMIDSHIFTED+(dc_yl-centery);
						dest = ylookup[dc_yl]+columnofs[x];
						do
						{
							*dest = *source++;
							dest += SCREENWIDTH;
						} while(count--);
*/
					}
				}
				continue; // Next visplane
			}
		}
		// Regular flat
		tempSource = W_CacheLumpNum(firstflat+
			flattranslation[pl->picnum], PU_STATIC);
		scrollOffset = leveltime>>1&63;
		switch(pl->special)
		{ // Handle scrolling flats
			case 201: case 202: case 203: // Scroll_North_xxx
				ds_source = tempSource+((scrollOffset
					<<(pl->special-201)&63)<<6);
				break;
			case 204: case 205: case 206: // Scroll_East_xxx
				ds_source = tempSource+((63-scrollOffset)
					<<(pl->special-204)&63);
				break;
			case 207: case 208: case 209: // Scroll_South_xxx
				ds_source = tempSource+(((63-scrollOffset)
					<<(pl->special-207)&63)<<6);
				break;
			case 210: case 211: case 212: // Scroll_West_xxx
				ds_source = tempSource+(scrollOffset
					<<(pl->special-210)&63);
				break;
			case 213: case 214: case 215: // Scroll_NorthWest_xxx
				ds_source = tempSource+(scrollOffset
					<<(pl->special-213)&63)+((scrollOffset
					<<(pl->special-213)&63)<<6);
				break;
			case 216: case 217: case 218: // Scroll_NorthEast_xxx
				ds_source = tempSource+((63-scrollOffset)
					<<(pl->special-216)&63)+((scrollOffset
					<<(pl->special-216)&63)<<6);
				break;
			case 219: case 220: case 221: // Scroll_SouthEast_xxx
				ds_source = tempSource+((63-scrollOffset)
					<<(pl->special-219)&63)+(((63-scrollOffset)
					<<(pl->special-219)&63)<<6);
				break;
			case 222: case 223: case 224: // Scroll_SouthWest_xxx
				ds_source = tempSource+(scrollOffset
					<<(pl->special-222)&63)+(((63-scrollOffset)
					<<(pl->special-222)&63)<<6);
				break;
			default:
				ds_source = tempSource;
				break;
		}
		planeheight = abs(pl->height-viewz);
		light = (pl->lightlevel >> LIGHTSEGSHIFT)+extralight;
		if(light >= LIGHTLEVELS)
		{
			light = LIGHTLEVELS-1;
		}
		if(light < 0)
		{
			light = 0;
		}
		planezlight = zlight[light];

		pl->top[pl->maxx+1] = 0xffff;
		pl->top[pl->minx-1] = 0xffff;

		stop = pl->maxx+1;
		for(x = pl->minx; x <= stop; x++)
		{
			R_MakeSpans(x, pl->top[x-1], pl->bottom[x-1],
				pl->top[x], pl->bottom[x]);
		}
		Z_ChangeTag(tempSource, PU_CACHE);
	}
}
